/*
 *  Roadnav
 *  DirectionsFrame.cpp
 *
 *  Copyright (c) 2004 - 2007 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// This frame is responsible for showing and verbalizing real time
/// directions to a place the user specifies.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#include <wx/wx.h>

#include "DirectionsFrame.h"
#include "Frame.h"
#include "Support.h"
#include "libroadnav/Map.h"
#include "libroadnav/MapRepresentations.h"
#include "libroadnav/MapControlDataImporter_GNISDECI.h"
#include "libroadnav/MapControlDataImporter_TigerLine.h"
#include "libroadnav/MapState.h"
#include "libroadnav/Debug.h"
#include "libroadnav/MapSupport.h"
#include "TTS.h"
#include "App.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// DirectionsFrame event table
///
//////////////////////////////////////////////////////////////////////////////
BEGIN_EVENT_TABLE(DirectionsFrame, wxFrame)
	EVT_CLOSE(DirectionsFrame::OnClose)
END_EVENT_TABLE()

//////////////////////////////////////////////////////////////////////////////
///
/// \brief DirectionsFrame construct - create children, set up timer
///
/////////////////////////////////////////////////////////////////////////////
DirectionsFrame::DirectionsFrame (
		class MapFrame * pfrmMap
	)
	: 
	wxFrame
	(
		(wxFrame *) pfrmMap,
		-1,
		wxT("Directions"),
		wxDefaultPosition,
		wxDefaultSize,
		wxDEFAULT_FRAME_STYLE
	),
	m_fntText(16, wxDEFAULT, wxNORMAL, wxNORMAL)
{
	wxBoxSizer * psizerWnd;
	
	m_pfrmMap = pfrmMap;
	
	m_bLockOutGPSUpdate = false;
	m_bLiveMode = false;
	m_bGPSLock = false;
	m_fGPSHeading = 0;
	m_fGPSSpeed = 0;
	
	psizerWnd = new wxBoxSizer(wxVERTICAL);
	
	//////////////////////////////////////////////////////////////////////////
	// Label
	//////////////////////////////////////////////////////////////////////////
	m_pctlText = new wxTextCtrl(this, -1, wxT(""), wxDefaultPosition, wxSize(300, 100), wxTE_MULTILINE | wxTE_RICH),

	psizerWnd->Add(
		m_pctlText,
		1,
		wxGROW,
		0
	);

	wxTextAttr taEdit(m_clrText, m_clrBackground, m_fntText);

	m_pctlText->SetDefaultStyle(taEdit);

	//////////////////////////////////////////////////////////////////////////
	// Set up the sizer
	//////////////////////////////////////////////////////////////////////////

	//psizerWnd->Fit(this);
	SetSizer(psizerWnd);
	Layout();
	psizerWnd->SetSizeHints(this);

	m_pcRecords = new MapControlData_Tiles;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief DirectionsFrame destructor - doesn't do anything
///
//////////////////////////////////////////////////////////////////////////////
DirectionsFrame::~DirectionsFrame()
{
	delete m_pcRecords;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief User wants to bring up directions window - get destination address 
/// and make sure everything is okay before letting the window appear.
/// 
/// If there's no GPS lock, then get the starting address too.
///
//////////////////////////////////////////////////////////////////////////////
bool DirectionsFrame::Setup()
{
	wxFont fntThis = GetFont();

	MapMarker cStartMarker;
	MapMarker cEndMarker;
	
	wxString strErrorMsg;
	Address sEnd;
	
	//////////////////////////////////////////////////////////////////////////
	// GPS lock required first
	//////////////////////////////////////////////////////////////////////////

	m_bLiveMode = m_bGPSLock;

	//////////////////////////////////////////////////////////////////////////
	// Get starting address
	//////////////////////////////////////////////////////////////////////////
	if (m_bLiveMode)
	{
		if (!m_bGPSLock)
		{
			// obviously, this can't happen right now .. this is more for future purposes
			wxMessageDialog dlgError(this, wxT("GPS lock required for live directions"), wxT("Error"), wxOK | wxICON_ERROR);
			dlgError.ShowModal();
			return true;
		}
	}
	else
	{
		if (QueryAddress(	wxT("Enter the starting address"),
							m_pfrmMap,
							m_pfrmMap->GetMapControl(),
							wxT(""),
							wxT(""),
							wxT(""),
							true,
							&cStartMarker,
							&strErrorMsg) != wxID_OK)
		{
			return true;
		}
		
		if (!cStartMarker.m_cAddress.m_bValid)
		{
			wxMessageDialog dlgError(this, strErrorMsg, wxT("Error"), wxOK | wxICON_ERROR);
			return true;
		}
		
		if (cStartMarker.m_strName == wxT("Selected Address") || cStartMarker.m_strName == wxT(""))
			cStartMarker.m_strName = wxT("Starting Address");
			
		m_sStart = cStartMarker.m_cAddress;
	}

	//////////////////////////////////////////////////////////////////////////
	// Get destination address
	//////////////////////////////////////////////////////////////////////////

	if (QueryAddress(	wxT("Enter the destination address"),
						m_pfrmMap,
						m_pfrmMap->GetMapControl(),
						wxT(""),
						wxT(""),
						wxT(""),
						true,
						&cEndMarker,
						&strErrorMsg) != wxID_OK)
	{
		return true;
	}
		
	if (!cEndMarker.m_cAddress.m_bValid)
	{
		wxMessageDialog dlgError(this, strErrorMsg, wxT("Error"), wxOK | wxICON_ERROR);
		return true;
	}

	if (cEndMarker.m_strName == wxT("Selected Address") || cEndMarker.m_strName == wxT(""))
		cEndMarker.m_strName = wxT("Destination Address");

	m_sEnd = cEndMarker.m_cAddress;

	//////////////////////////////////////////////////////////////////////////
	// Is it possible to get there at all?
	//////////////////////////////////////////////////////////////////////////
	if (ComputePath())
	{
		wxMessageDialog dlgError(this, wxT("Unable to determine route!"), wxT("Error"), wxOK | wxICON_ERROR);
		dlgError.ShowModal();
		return true;
	}

	//////////////////////////////////////////////////////////////////////////
	// write out directions
	//////////////////////////////////////////////////////////////////////////
	m_pctlText->SetValue(m_pcRecords->TranslateRecordsToDirections(m_vrPath));

	m_strLastNext = wxT("");

	//////////////////////////////////////////////////////////////////////////
	// update font
	//////////////////////////////////////////////////////////////////////////
	wxTextAttr taEdit(m_clrText, m_clrBackground, m_fntText);
	m_pctlText->SetDefaultStyle(taEdit);
	m_pctlText->SetStyle(0, m_pctlText->GetLastPosition(), taEdit);

	if (!m_bLiveMode)
	{
		m_pfrmMap->GetMapControl()->SetMarker(cStartMarker);
	}

	m_pfrmMap->GetMapControl()->SetMarker(cEndMarker);

	m_pfrmMap->GetMapControl()->SetCenterCoordinates((m_sStart.m_ptCoordinates + m_sEnd.m_ptCoordinates) / 2);
	m_pfrmMap->GetMapControl()->SetDetailLevel( m_pfrmMap->GetMapControlData().GetNumRecordTypes()-1 );
	m_pfrmMap->GetMapControl()->AutoScale();
	
	m_bLockOutGPSUpdate = false;

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Compute the shortest path from m_ptGPS to m_sEnd
///
/// Updates m_sStart (address of m_ptGPS), m_vrPath (path in record numbers),
/// m_vptPath (path in long/lat coordinates) in the process
///
//////////////////////////////////////////////////////////////////////////////
bool DirectionsFrame::ComputePath()
{
	int iAutoDownloadMaps = 1;
	
	g_pConfig->Read(wxT("AutoDownloadMaps"), &iAutoDownloadMaps, 1);

	LibRoadnavDebug2(wxT("Routing"), wxT("ComputePath called - iAutoDownloadMaps = %d, m_bLiveMode = %d"), iAutoDownloadMaps, m_bLiveMode);

	// determine the path

	if (m_bLiveMode)
	{
		LibRoadnavDebug2(wxT("Routing"), wxT("Trying FindCoordinates for %.7f, %.7f"), m_ptGPS.GetLong(), m_ptGPS.GetLat());

		m_sStart = FindCoordinates(m_ptGPS, OnlyRoadsVisibility());

		LibRoadnavDebug1(wxT("Routing"), wxT("m_sStart.m_bValid = %d"), m_sStart.m_bValid);

		if (!m_sStart.m_bValid)
			return true;
	}

	if (iAutoDownloadMaps)
	{
		bool bVisibility[LAST_RECORD_TYPE + 1];
		int iRec;
		int iCount;
		map<int,bool> mapCounties;
		map<int,bool>::iterator iCounties;
		int iRT;
		double fMargin = 1;

		// only load county boundaries
		for (iRT = FIRST_RECORD_TYPE; iRT <= LAST_RECORD_TYPE; iRT++)
			bVisibility[iRT] = (iRT == RecordTypeCountyBoundary);

		// calculate top-left and bottom-right coordinates
		Point ptTopLeft;
		Point ptBottomRight;
		
		ptTopLeft = m_sStart.m_ptCoordinates;
		ptBottomRight = m_sEnd.m_ptCoordinates;
		
		if (ptTopLeft.GetLongRaw() > ptBottomRight.GetLongRaw())
		{
			wxInt32 iTmp;
			
			iTmp = ptTopLeft.GetLongRaw();
			ptTopLeft.SetLongRaw(ptBottomRight.GetLongRaw());
			ptBottomRight.SetLongRaw(iTmp);
		}

		if (ptTopLeft.GetLatRaw() > ptBottomRight.GetLatRaw())
		{
			wxInt32 iTmp;
			
			iTmp = ptTopLeft.GetLatRaw();
			ptTopLeft.SetLatRaw(ptBottomRight.GetLatRaw());
			ptBottomRight.SetLatRaw(iTmp);
		}
		
		ptTopLeft = ptTopLeft - Point(fMargin, fMargin);
		ptBottomRight = ptBottomRight + Point(fMargin, fMargin);

		// load the county boundaries for the region that the route may cover
		m_pcRecords->LoadRegion(ptTopLeft, ptBottomRight, bVisibility);
		
		// enumerate counties present
		iCount = m_pcRecords->Count();
		for (iRec = 0; iRec < iCount; iRec++)
		{
			IMapControlDataEntry * pcRec;
			Point * pptRec;
			int iCoordinate;
			
			pcRec = m_pcRecords->GetRecord(iRec);
			
			for (iCoordinate = 0; iCoordinate < pcRec->nCoordinates; iCoordinate++)
			{
				pptRec = &pcRec->psCoordinates[iCoordinate];
				
				if (pptRec->GetLongRaw() >= ptTopLeft.GetLongRaw() &&
					pptRec->GetLongRaw() <= ptBottomRight.GetLongRaw() &&
					pptRec->GetLatRaw() >= ptTopLeft.GetLatRaw() &&
					pptRec->GetLatRaw() <= ptBottomRight.GetLatRaw())
				{
					mapCounties[pcRec->CountyCode()] = true;
				}
			}
		}

		map<int,bool> bDownloadingCounty;
		map<wxString,bool> bDownloadingState;
		vector<wxInt32> vDownloadCounties;
		vector<wxString> vDownloadStates;
		int iTotal;

		// compile list of what's needed
		iTotal = 0;
		for (iCounties = mapCounties.begin(); iCounties != mapCounties.end(); iCounties++)
		{
			MapControlDataImporter_GNISDECI diGNIS;
			MapControlDataImporter_TigerLine diTigerLine;
			wxString strState;
			int iCountyCode;
			
			iCountyCode = (*iCounties).first;
			strState = StateAbbreviationByCode(iCountyCode / 1000);
			
			// check county code
			if (iCountyCode > 0)
			{
				wxASSERT(strState != wxT(""));

				if (!diTigerLine.IsCountyLoaded(iCountyCode) && bDownloadingCounty.find(iCountyCode) == bDownloadingCounty.end())
				{
					bDownloadingCounty[iCountyCode] = true;
					vDownloadCounties.push_back(iCountyCode);
				}

				if (!diGNIS.IsGNISStateLoaded(strState) && bDownloadingState.find(strState) == bDownloadingState.end())
				{
					bDownloadingState[strState] = true;
					vDownloadStates.push_back(strState);
				}
			}
		}

		iTotal = bDownloadingCounty.size() + bDownloadingState.size();

		LibRoadnavDebug2(wxT("Routing"), wxT("Want to download %d counties and %d states"), bDownloadingCounty.size(), bDownloadingState.size());
		
		// does anything need to be downloaded?		
		if (iTotal)
		{
			MapControlDataImporter_TigerLine diTigerLine;
			MapControlDataImporter_GNISDECI diGNIS;

			ProgressDialog dlgProgress(NULL, wxT("Downloading..."), iTotal * 100);
			dlgProgress.Show();
			
			diTigerLine.DownloadCounties(m_pcRecords, vDownloadCounties, &dlgProgress, 0, vDownloadCounties.size() * 100);
			diGNIS.DownloadGNISStates(m_pcRecords, vDownloadStates, &dlgProgress, vDownloadCounties.size() * 100, vDownloadStates.size() * 100);
		}

		LibRoadnavDebug0(wxT("Routing"), wxT("Downloading complete"));
	}

	LibRoadnavDebug4(wxT("Routing"), wxT("Calculating route from %.7f, %.7f to %.7f to %.7f"), 
														m_sStart.m_ptCoordinates.GetLong(),
														m_sStart.m_ptCoordinates.GetLat(),
														m_sEnd.m_ptCoordinates.GetLong(),
														m_sEnd.m_ptCoordinates.GetLat()
														);
														
	m_vrPath = m_pcRecords->ShortestPath(m_sStart.m_ptCoordinates, m_sEnd.m_ptCoordinates, m_sStart.m_idRecord, m_sEnd.m_idRecord);

	if (!m_vrPath.size())
		return true;

	LibRoadnavDebug0(wxT("Routing"), wxT("Shortest route found"));
	LibRoadnavDebug0(wxT("Routing"), wxT("Translating records to coordinates"));

	/// Current path in long/lat
	vector<Point> vptPath;
	vptPath = m_pcRecords->TranslateRecordsToCoordinates(m_vrPath);

	LibRoadnavDebug0(wxT("Routing"), wxT("Translation done"));

	MapTrack cTrack(wxT("Directions"), vptPath, wxColour(255, 0, 0), true);

	// plot
	LibRoadnavDebug0(wxT("Routing"), wxT("Plotting track"));
	m_pfrmMap->GetMapControl()->SetTrack(cTrack);
	LibRoadnavDebug0(wxT("Routing"), wxT("Done plotting"));

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Updated GPS coordinates received
///
/// Record the coordinates, figure out if we're on path, give any verbal cues
/// necessary, and recompute the path if the user is off course
///
//////////////////////////////////////////////////////////////////////////////
void DirectionsFrame::OnGPSUpdate(bool bEnabled, bool bActive, bool bLocked, Point pt, double fSpeed, double fHeading, wxString strLockType, int nSat, Address adrGPS)
{
	unsigned int i;
	double fBestDistanceAbs = INFINITY;
	double fBestDistanceMiles = INFINITY;
	int iBestLeg = -1;
	vector<tRecordID> vrPathLeft;
	wxString strNext;
	static MapControlData_Tiles cRecLocal;
	Address adrGPSLocal;
	
	cRecLocal.LoadRegion(pt, pt, OnlyRoadsVisibility());
	adrGPSLocal = FindCoordinates(pt, OnlyRoadsVisibility(), false, true, &cRecLocal);

	LibRoadnavDebug0(wxT("Routing"), wxT("OnGPSUpdate"));

	//////////////////////////////////////////////////////////////////////////
	// remember the GPS state
	//////////////////////////////////////////////////////////////////////////
	m_bGPSLock = bLocked;
	m_ptGPS = pt;
	m_fGPSHeading = fHeading;
	m_fGPSSpeed = fSpeed;

	//////////////////////////////////////////////////////////////////////////
	// don't do anything if the window isn't visible
	//////////////////////////////////////////////////////////////////////////
	if (!IsShown())
	{
		LibRoadnavDebug0(wxT("Routing"), wxT("Not visible - OnGPSUpdate exiting"));
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// If live mode is off, then don't bother with the calculations either
	//////////////////////////////////////////////////////////////////////////
	if (!m_bLiveMode)
	{
		LibRoadnavDebug0(wxT("Routing"), wxT("Live mode off - OnGPSUpdate exiting"));
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// lost GPS lock? stop now then
	//////////////////////////////////////////////////////////////////////////
	if (!bLocked)
	{
		LibRoadnavDebug0(wxT("Routing"), wxT("No lock - OnGPSUpdate exiting"));
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Lockout?
	//////////////////////////////////////////////////////////////////////////
	if (m_bLockOutGPSUpdate)
	{
		LibRoadnavDebug0(wxT("Routing"), wxT("Lockout - OnGPSUpdate exiting"));
		return;
	}
	m_bLockOutGPSUpdate = true;
	
	//////////////////////////////////////////////////////////////////////////
	// find the leg of the trip that we're closest too
	//////////////////////////////////////////////////////////////////////////
	LibRoadnavDebug0(wxT("Routing"), wxT("Finding closest leg of route"));
	for (i = 0; i < m_vrPath.size(); i++)
	{
		double fDistance;
		IMapControlDataEntry * psRec;
		int iCoordinate;

		psRec = m_pcRecords->GetRecordByID(m_vrPath[i]);
		
		wxASSERT(psRec);

		for (iCoordinate = 0; iCoordinate < psRec->nCoordinates; iCoordinate++)
		{
			fDistance = (psRec->psCoordinates[iCoordinate] - adrGPSLocal.m_ptCoordinates).abs();

			if (fDistance < fBestDistanceAbs)
			{
				fBestDistanceAbs = fDistance;
				fBestDistanceMiles = Distance(psRec->psCoordinates[iCoordinate], adrGPSLocal.m_ptCoordinates);
				iBestLeg = i;
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// if more than 0.1 miles from path, recompute path
	//////////////////////////////////////////////////////////////////////////
	if (fBestDistanceMiles > 0.1)
	{
		LibRoadnavDebug0(wxT("Routing"), wxT("Too far from route - recomputing"));

		int iLiveDirectionsTTS = 1;
		g_pConfig->Read(wxT("LiveDirectionsTTS"), &iLiveDirectionsTTS, 1);

		if (iLiveDirectionsTTS)
			m_pfrmMap->GetTTS()->Speak(wxT("Recalculating path"));

		ComputePath();
		iBestLeg = 0;
	}

	//////////////////////////////////////////////////////////////////////////
	// ignore the part of the path that has been covered and just show
	// what's left
	//////////////////////////////////////////////////////////////////////////
	LibRoadnavDebug0(wxT("Routing"), wxT("Figuring out what's left"));
	for (i = iBestLeg; i < m_vrPath.size(); i++)
		vrPathLeft.push_back(m_vrPath[i]);

	if (vrPathLeft.size() >= 2)
	{
		//////////////////////////////////////////////////////////////////////////
		// if there are at least two legs left, then figure out some directions
		//////////////////////////////////////////////////////////////////////////
		double fDeltaAngle = 0;
		wxString strDirectionChange = wxT("");
		vector<IMapControlDataEntry *> vAssociatedRecords;
		
		LibRoadnavDebug0(wxT("Routing"), wxT("Looking up intersection information"));

		vAssociatedRecords = m_pcRecords->IntersectionInformation(vrPathLeft[0], vrPathLeft[1], &fDeltaAngle, &strDirectionChange);

		if (strDirectionChange == wxT("continue"))
			strDirectionChange = wxT("go straight");

		if (strDirectionChange.Length() > 0)
			strDirectionChange.GetWritableChar(0) = toupper(strDirectionChange.GetChar(0));

		strNext = strDirectionChange + wxT(" onto ") + m_pcRecords->GetRecordByID(vrPathLeft[1])->NameAndType(0, true);

	}
	else
	{
		//////////////////////////////////////////////////////////////////////////
		// we must be at the destination if there's one leg left
		//////////////////////////////////////////////////////////////////////////
		strNext = wxT("At destination");
	}

	//////////////////////////////////////////////////////////////////////////
	// write out directions
	//////////////////////////////////////////////////////////////////////////
	
	wxString strDirections;
	
	LibRoadnavDebug0(wxT("Routing"), wxT("Writing out directions"));

	strDirections = m_pcRecords->TranslateRecordsToDirections(vrPathLeft);
	
	m_pctlText->SetValue(strNext + wxT("\r\n") + 
							strDirections + wxT("\r\n") +
							m_sEnd.FormatAddress()
							);

	//////////////////////////////////////////////////////////////////////////
	// update font
	//////////////////////////////////////////////////////////////////////////
	LibRoadnavDebug0(wxT("Routing"), wxT("Fixing UI"));
	wxTextAttr taEdit(m_clrText, m_clrBackground, m_fntText);
	m_pctlText->SetDefaultStyle(taEdit);
	m_pctlText->SetStyle(0, m_pctlText->GetLastPosition(), taEdit);

	//////////////////////////////////////////////////////////////////////////
	// verbalize if the current action is different than the previous action
	//////////////////////////////////////////////////////////////////////////
	if (strNext != m_strLastNext)
	{
		int iLiveDirectionsTTS = 1;
		g_pConfig->Read(wxT("LiveDirectionsTTS"), &iLiveDirectionsTTS, 1);

		if (iLiveDirectionsTTS)
		{
			LibRoadnavDebug0(wxT("Routing"), wxT("Speaking directions"));
			m_pfrmMap->GetTTS()->Speak(strNext);
			LibRoadnavDebug0(wxT("Routing"), wxT("Done speaking"));
		}

		m_pfrmMap->SetVariable(wxT("Directions - Next Step"), strNext);

		m_strLastNext = strNext;
	}

	wxString strBalance;

	LibRoadnavDebug0(wxT("Routing"), wxT("Updating skin variables"));
	
	strBalance = strDirections;
	m_pfrmMap->SetVariable(wxT("Directions - Line 1"), strBalance.BeforeFirst(wxT('\n')));
	
	strBalance = strBalance.AfterFirst(wxT('\n'));
	m_pfrmMap->SetVariable(wxT("Directions - Line 2"), strBalance.BeforeFirst(wxT('\n')));

	strBalance = strBalance.AfterFirst(wxT('\n'));
	m_pfrmMap->SetVariable(wxT("Directions - Line 3"), strBalance.BeforeFirst(wxT('\n')));

	strBalance = strBalance.AfterFirst(wxT('\n'));
	m_pfrmMap->SetVariable(wxT("Directions - Line 4"), strBalance.BeforeFirst(wxT('\n')));

	m_bLockOutGPSUpdate = false;

	LibRoadnavDebug0(wxT("Routing"), wxT("OnGPSUpdate done"));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief User wants to close this window - notify parent and hide.
///
//////////////////////////////////////////////////////////////////////////////
void DirectionsFrame::OnClose(wxCloseEvent& event)
{
	m_pfrmMap->GetMapControl()->ClearMarker(wxT("Starting Address"));
	m_pfrmMap->GetMapControl()->ClearMarker(wxT("Destination Address"));
	m_pfrmMap->GetMapControl()->ClearTrack(wxT("Directions"));
	Hide();
	m_pfrmMap->OnDirectionsWindowHiding();
	m_pcRecords->Clear();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Propagates SetBackgroundColour calls down to m_pctlText
///
//////////////////////////////////////////////////////////////////////////////
bool DirectionsFrame::SetBackgroundColour(const wxColour& colour)
{
	if (colour != m_clrBackground)
	{
		m_clrBackground = colour;
		
		wxTextAttr taEdit(m_clrText, m_clrBackground, m_fntText);
		m_pctlText->SetDefaultStyle(taEdit);
		m_pctlText->SetStyle(0, m_pctlText->GetLastPosition(), taEdit);
		m_pctlText->SetBackgroundColour(m_clrBackground);
	}
	
	return wxFrame::SetBackgroundColour(colour);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Propagates SetForegroundColour calls down to m_pctlText
///
//////////////////////////////////////////////////////////////////////////////
bool DirectionsFrame::SetForegroundColour(const wxColour& colour)
{
	if (colour != m_clrText)
	{
		m_clrText = colour;
		
		wxTextAttr taEdit(m_clrText, m_clrBackground, m_fntText);
		m_pctlText->SetDefaultStyle(taEdit);
		m_pctlText->SetStyle(0, m_pctlText->GetLastPosition(), taEdit);
	}
	
	return wxFrame::SetForegroundColour(colour);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Propagates SetFont calls down to m_pctlText
///
//////////////////////////////////////////////////////////////////////////////
bool DirectionsFrame::SetFont(const wxFont& font)
{
	m_fntText = font;

	wxTextAttr taEdit(m_clrText, m_clrBackground, m_fntText);
	m_pctlText->SetDefaultStyle(taEdit);
	m_pctlText->SetStyle(0, m_pctlText->GetLastPosition(), taEdit);
	
	return wxFrame::SetFont(font);
}

