/*
 *  Roadnav
 *  EditThemeDialog.cpp
 *
 *  Copyright (c) 2004 - 2007 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains a dialog box used to edit a specific theme.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#include <wx/wx.h>

#include "EditThemeDialog.h"

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// IDs for EditThemeDialog.cpp
///
//////////////////////////////////////////////////////////////////////////////
enum
{
	idName,
	idDetailLevel,
	idRecordType,
	idColor
};

//////////////////////////////////////////////////////////////////////////////
///
/// EditThemeDialog event table
///
//////////////////////////////////////////////////////////////////////////////
BEGIN_EVENT_TABLE(EditThemeDialog, wxDialog)
	EVT_TEXT(idDetailLevel, EditThemeDialog::OnDetailLevelChange)
	EVT_TEXT(idRecordType, EditThemeDialog::OnRecordTypeChange)
	EVT_BUTTON(wxID_OK, EditThemeDialog::OnOk)
	EVT_CUSTOM_RANGE(wxEVT_COMMAND_CHECKBOX_CLICKED, 0, EDITTHEMEDIALOG_MAX_DETAILLEVELS * 2, EditThemeDialog::OnCheckBox)
END_EVENT_TABLE()

//////////////////////////////////////////////////////////////////////////////
///
/// \brief EditThemeDialog constructor - create and initialize controls
///
//////////////////////////////////////////////////////////////////////////////
EditThemeDialog::EditThemeDialog(
									wxWindow * pParent, 
									const wxString & strTheme, 
									const MapAppearanceSettings & cTheme, 
									IMapControlData * pMapControlData)
										: wxDialog(	pParent, 
													-1, 
													wxT("Edit Theme"), 
													wxDefaultPosition, 
													wxDefaultSize),
										m_Theme(cTheme), 
										m_strTheme(strTheme)
{
	wxBoxSizer * psizerWnd;
	wxFlexGridSizer * psizerGrid;
	int iDetailLevel;
	int iRow;
	int iCol;
	int nDetailLevels;
	int iHeading;
	wxChar * pszHeadings[] = {wxT("Detail Lvl"), wxT(""), wxT("Clr"), wxT("Font Size"), wxT(""), wxT("Clr"), wxT("Line Width"), NULL};

	m_pMapControlData = pMapControlData,
	
	nDetailLevels = m_Theme.GetDetailLevelCount();
	
	wxASSERT(nDetailLevels <= EDITTHEMEDIALOG_MAX_DETAILLEVELS);
	
	//////////////////////////////////////////////////////////////////////////
	// Box sizer for entire window
	//////////////////////////////////////////////////////////////////////////
	psizerWnd = new wxBoxSizer(wxVERTICAL);

	//////////////////////////////////////////////////////////////////////////
	// Sizer for top half
	//////////////////////////////////////////////////////////////////////////
	psizerGrid = new wxFlexGridSizer(2, 4, 10);

	psizerWnd->Add(
		psizerGrid,
		0,
		wxGROW | wxTOP | wxLEFT | wxRIGHT,
		10
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Theme Name
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, idName, wxT("Theme Name")),
		1,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Theme Name Control
	//////////////////////////////////////////////////////////////////////////
	m_pctlThemeName = new wxTextCtrl(this, idName, m_strTheme);

	psizerGrid->Add(m_pctlThemeName, 
				1, 
				wxGROW, 
				0);
	
	//////////////////////////////////////////////////////////////////////////
	// Feature type text
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Feature Type")),
		1,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Feature type combo box
	//////////////////////////////////////////////////////////////////////////
	m_pctlRecordType = new wxComboBox(this, idRecordType, wxT(""), wxDefaultPosition, wxSize(200, -1), 0, NULL, wxCB_READONLY);
	m_pctlRecordType->Append(GetRecordTypeNames());	
	wxASSERT(m_pctlRecordType->GetCount() >= 1);
	m_pctlRecordType->SetSelection(0);
	
	psizerGrid->Add(m_pctlRecordType,
				1,
				wxGROW,
				0
	);

	//////////////////////////////////////////////////////////////////////////
	// Static box surrounding feature specific settings
	//////////////////////////////////////////////////////////////////////////
	wxStaticBoxSizer * psizerDetailSpecific = new wxStaticBoxSizer(wxVERTICAL, this, wxT("Feature Settings"));

	psizerWnd->Add(psizerDetailSpecific,
				1,
				wxGROW | wxLEFT | wxRIGHT,
				10
	);

	//////////////////////////////////////////////////////////////////////////
	// Flex grid inside static box
	//////////////////////////////////////////////////////////////////////////
	wxFlexGridSizer * psizerDetailSpecificGrid = new wxFlexGridSizer((sizeof(pszHeadings) / sizeof(wxChar *) - 1) * 2, 1, 6);

	psizerDetailSpecific->Add(psizerDetailSpecificGrid,
				1,
				wxGROW,
				0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Column headings for flex grid inside static box
	//////////////////////////////////////////////////////////////////////////
	for (iCol = 0; iCol < 2; iCol++)
	{
		for (iHeading = 0; pszHeadings[iHeading]; iHeading++)
		{
			psizerDetailSpecificGrid->Add(
				new wxStaticText(this, -1, pszHeadings[iHeading]),
				1,
				wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT | wxLEFT,
				5
			);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Go through each detail level
	//////////////////////////////////////////////////////////////////////////
	for (iRow = 0; iRow < (nDetailLevels + 1) / 2; iRow++)
	{
		for (iCol = 0; iCol < 2; iCol++)
		{
			iDetailLevel = iRow + iCol * ((nDetailLevels + 1) / 2);
			
			if (iDetailLevel < nDetailLevels)
			{
				//////////////////////////////////////////////////////////////////////////
				// Detail level label
				//////////////////////////////////////////////////////////////////////////
				psizerDetailSpecificGrid->Add(
					new wxStaticText(this, -1, wxString::Format(wxT("%d"), iDetailLevel)),
					1,
					wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
					0
				);
			

				//////////////////////////////////////////////////////////////////////////
				// Label Visible checkbox
				//////////////////////////////////////////////////////////////////////////
				m_pctlLabelVisible[iDetailLevel] = new wxCheckBox(this, iDetailLevel, wxT(""));
				
				psizerDetailSpecificGrid->Add(m_pctlLabelVisible[iDetailLevel],
							1,
							wxALIGN_RIGHT | wxLEFT,
							10
				);

				//////////////////////////////////////////////////////////////////////////
				// Label color control
				//////////////////////////////////////////////////////////////////////////
				m_pctlLabelColor[iDetailLevel] = new ColorPickerControl(this, idColor, wxDefaultPosition, wxSize(30, 20));
				
				psizerDetailSpecificGrid->Add(m_pctlLabelColor[iDetailLevel],
							1,
							wxFIXED_MINSIZE | wxALIGN_RIGHT,
							0
				);

				//////////////////////////////////////////////////////////////////////////
				// Label Font Size text control
				//////////////////////////////////////////////////////////////////////////
				m_pctlLabelFontSize[iDetailLevel] = new wxTextCtrl(this, -1, wxT(""));
				
				psizerDetailSpecificGrid->Add(m_pctlLabelFontSize[iDetailLevel],
							1,
							wxGROW,
							0
				);

				//////////////////////////////////////////////////////////////////////////
				// Line Visible checkbox
				//////////////////////////////////////////////////////////////////////////

				m_pctlLineVisible[iDetailLevel] = new wxCheckBox(this, iDetailLevel + EDITTHEMEDIALOG_MAX_DETAILLEVELS, wxT(""));
				
				psizerDetailSpecificGrid->Add(m_pctlLineVisible[iDetailLevel],
							1,
							wxALIGN_RIGHT | wxLEFT,
							10
				);

				//////////////////////////////////////////////////////////////////////////
				// Line color control
				//////////////////////////////////////////////////////////////////////////
				m_pctlLineColor[iDetailLevel] = new ColorPickerControl(this, idColor, wxDefaultPosition, wxSize(30, 20));
				
				psizerDetailSpecificGrid->Add(m_pctlLineColor[iDetailLevel],
							1,
							wxFIXED_MINSIZE | wxALIGN_RIGHT,
							0
				);

				//////////////////////////////////////////////////////////////////////////
				// Line Width text control
				//////////////////////////////////////////////////////////////////////////
				m_pctlLineWidth[iDetailLevel] = new wxTextCtrl(this, -1, wxT(""));

				psizerDetailSpecificGrid->Add(m_pctlLineWidth[iDetailLevel],
							1,
							wxGROW,
							0
				);
			}
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// ok and cancel button
	//////////////////////////////////////////////////////////////////////////
	wxStdDialogButtonSizer* psizerButton = CreateStdDialogButtonSizer(wxOK | wxCANCEL);
	psizerWnd->Add(psizerButton, 0, wxALL | wxALIGN_CENTER, 5);
	
	//////////////////////////////////////////////////////////////////////////
	// Set up the sizer
	//////////////////////////////////////////////////////////////////////////
	m_pctlRecordType->SetFocus();
	psizerWnd->Fit(this);
	SetSizer(psizerWnd);
	Layout();
	psizerWnd->SetSizeHints(this);
	
	m_iLastRecordType = -1;
	
	wxCommandEvent ev;
	OnRecordTypeChange(ev);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Ok button pressed - save settings
///
//////////////////////////////////////////////////////////////////////////////
void EditThemeDialog::OnOk(wxCommandEvent& event)
{
	RecordSettings();

	EndModal(wxID_OK);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Save the active settings to m_Theme
///
//////////////////////////////////////////////////////////////////////////////
void EditThemeDialog::RecordSettings()
{
	m_strTheme = m_pctlThemeName->GetValue();

	if (m_iLastRecordType < 0)
		return;

	if (m_iLastRecordType <= LAST_MISCCOLOR )
		m_Theme.SetMiscColor((EMiscColor) m_iLastRecordType, m_pctlLabelColor[0]->GetColor());
	else
	{
		int iDetailLevel;
		
		ERecordType eRecordType = (ERecordType) (m_iLastRecordType - LAST_MISCCOLOR - 1);
		
		for (iDetailLevel = 0; iDetailLevel < m_Theme.GetDetailLevelCount(); iDetailLevel++)
		{
			long lTmp;
			
			m_Theme.SetLineColor(eRecordType, m_pctlLineColor[iDetailLevel]->GetColor(), iDetailLevel);
			m_Theme.SetTextColor(eRecordType, m_pctlLabelColor[iDetailLevel]->GetColor(), iDetailLevel);
		
			m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_bTextVisible = m_pctlLabelVisible[iDetailLevel]->GetValue();
			
			m_pctlLabelFontSize[iDetailLevel]->GetValue().ToLong(&lTmp);
			m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_iFontSize = lTmp;
			
			m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_bLineVisible = m_pctlLineVisible[iDetailLevel]->GetValue();
			
			m_pctlLineWidth[iDetailLevel]->GetValue().ToLong(&lTmp);
			m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_iWidth = lTmp;
		}
	}
	
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief User changed feature types
///
//////////////////////////////////////////////////////////////////////////////
void EditThemeDialog::OnRecordTypeChange(wxCommandEvent & event)
{
	bool bIsMiscColor;
	int iDetailLevel;

	RecordSettings();
	
	m_iLastRecordType = m_pctlRecordType->GetSelection();
	
	if (m_iLastRecordType < 0)
		return;

	bIsMiscColor = m_iLastRecordType <= LAST_MISCCOLOR;

	for (iDetailLevel = 0; iDetailLevel < m_Theme.GetDetailLevelCount(); iDetailLevel++)
	{
		if (iDetailLevel > 0)
		{
			m_pctlLineColor[iDetailLevel]->Enable(!bIsMiscColor);
			m_pctlLabelColor[iDetailLevel]->Enable(!bIsMiscColor);
		}
		else
		{
			m_pctlLineColor[iDetailLevel]->Enable(!bIsMiscColor);
			m_pctlLabelColor[iDetailLevel]->Enable(true);
		}
		
		m_pctlLabelVisible[iDetailLevel]->Enable(!bIsMiscColor);
		m_pctlLabelFontSize[iDetailLevel]->Enable(!bIsMiscColor);
		m_pctlLineVisible[iDetailLevel]->Enable(!bIsMiscColor);
		m_pctlLineWidth[iDetailLevel]->Enable(!bIsMiscColor);
	}

	if (bIsMiscColor)
	{
		m_pctlLabelColor[0]->SetColor(m_Theme.GetMiscColor((EMiscColor) m_iLastRecordType));
		m_pctlLineColor[0]->SetColor(*wxBLACK);

		for (iDetailLevel = 1; iDetailLevel < m_Theme.GetDetailLevelCount(); iDetailLevel++)
		{
			m_pctlLabelColor[iDetailLevel]->SetColor(*wxBLACK);
			m_pctlLineColor[iDetailLevel]->SetColor(*wxBLACK);
		}
	}
	else
	{
		ERecordType eRecordType = (ERecordType) (m_iLastRecordType - LAST_MISCCOLOR - 1);

		for (iDetailLevel = 0; iDetailLevel < m_Theme.GetDetailLevelCount(); iDetailLevel++)
		{
			bool bLabelVisible = m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_bTextVisible;
			bool bLineVisible = m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_bLineVisible;
			
			m_pctlLineColor[iDetailLevel]->SetColor(m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_clrLine);
			m_pctlLabelColor[iDetailLevel]->SetColor(m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_clrText);

			m_pctlLabelVisible[iDetailLevel]->SetValue(bLabelVisible);
			m_pctlLabelFontSize[iDetailLevel]->SetValue(wxString::Format(wxT("%d"), m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_iFontSize));
			m_pctlLabelFontSize[iDetailLevel]->Enable(bLabelVisible);

			m_pctlLineVisible[iDetailLevel]->SetValue(bLineVisible);
			m_pctlLineWidth[iDetailLevel]->SetValue(wxString::Format(wxT("%d"), m_Theme.GetDetailSettings(iDetailLevel, eRecordType).m_iWidth));
			m_pctlLineWidth[iDetailLevel]->Enable(bLineVisible);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief User changed detail level
///
//////////////////////////////////////////////////////////////////////////////
void EditThemeDialog::OnDetailLevelChange(wxCommandEvent & event)
{
	OnRecordTypeChange(event);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Creates a wxArrayString of record type names
///
//////////////////////////////////////////////////////////////////////////////
wxArrayString EditThemeDialog::GetRecordTypeNames()
{
	wxArrayString strNames;
	wxInt16 nRecordTypes = m_pMapControlData->GetNumRecordTypes();
	wxInt16 iRecordType;
	
	strNames.Add(wxT("Background"));
	strNames.Add(wxT("Foreground"));
	strNames.Add(wxT("Marker Background"));
	strNames.Add(wxT("Marker Foreground"));
	
	for (iRecordType = 0; iRecordType < nRecordTypes; iRecordType++)
		strNames.Add( m_pMapControlData->GetRecordTypeDisplayName(iRecordType) );
	
	return strNames;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the theme object
///
//////////////////////////////////////////////////////////////////////////////
MapAppearanceSettings EditThemeDialog::GetTheme()
{
	return m_Theme;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns the theme name
///
//////////////////////////////////////////////////////////////////////////////
wxString EditThemeDialog::GetThemeName()
{
   return m_strTheme;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Triggered when a checkbox is selected or deselected
///
/// Enable/disable the appropriate text box
///
//////////////////////////////////////////////////////////////////////////////
void EditThemeDialog::OnCheckBox(wxEvent & event)
{
	if (event.GetId() < EDITTHEMEDIALOG_MAX_DETAILLEVELS)
		m_pctlLabelFontSize[ event.GetId() ]->Enable( m_pctlLabelVisible[event.GetId()]->GetValue() );
	else if (event.GetId() < 2 * EDITTHEMEDIALOG_MAX_DETAILLEVELS)
		m_pctlLineWidth[ event.GetId() - EDITTHEMEDIALOG_MAX_DETAILLEVELS ]->Enable( m_pctlLineVisible[event.GetId() - EDITTHEMEDIALOG_MAX_DETAILLEVELS]->GetValue() );
	else
		wxASSERT(0);
}
