/*
 *  Roadnav
 *  GPSInterface_Serial.cpp
 *
 *  Copyright (c) 2004 - 2007 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// NMEA serial gps interface module.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#include <wx/wx.h>

#include "libroadnav/Debug.h"

#include "App.h"
#include "GPSInterface_Serial.h"
#include "Support.h"

#ifdef ROADNAV_SERIAL_SUPPORT

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Constructor - zero out everything.
///
//////////////////////////////////////////////////////////////////////////////
GPSInterface_Serial::GPSInterface_Serial() : m_tLastGPSSentence((time_t) 0), m_tLastGPSLock((time_t) 0), m_tLastSerialActivity((time_t) 0)
{
	LibRoadnavDebug0(wxT("SerialIO"), wxT("Initializing serial I/O"));
	m_bSerialInitialized = false;
	
	m_fLastHeading = -1;

	m_fSpeed = 0;
	m_fHeading = 0;

	m_fMinimumSpeedForHeading = 5.0;

	m_bSeenGPGGA = false;
	m_bSeenGPRMC = false;

	m_nSatellites = 0;
	
	m_nSatellitesVisible = 0;
	
	*m_szBuffer = 0;

	m_strFixType = wxT("Invalid");
	
	m_pThread = NULL;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Destructor - shut down serial code
///
//////////////////////////////////////////////////////////////////////////////
GPSInterface_Serial::~GPSInterface_Serial()
{
	if (m_bSerialInitialized)
		ShutdownSerialIO();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns name of this GPS module.
///
//////////////////////////////////////////////////////////////////////////////
wxString GPSInterface_Serial::Name()
{
	return wxT("Serial");
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns if the checksum on a gps sentence is valid or not.
///
/// \param strSentence			GPS sentence. May be CR-LF terminated. Should
///								start with $.
/// \return						True if checksum valid, false otherwise.
///
//////////////////////////////////////////////////////////////////////////////
bool GPSInterface_Serial::VerifyGPSChecksum(wxString strSentence)
{
	unsigned int iPos;
	wxUint32 uiChecksum;
	
	uiChecksum = 0;
	
	LibRoadnavDebug1(wxT("SerialIO"), wxT("Verifying checksum on: %s"), strSentence.c_str());
	
	while (strSentence.Length() >= 3 && 
				(strSentence.GetChar(strSentence.Length() - 1) == wxT('\r') ||
				strSentence.GetChar(strSentence.Length() - 1) == wxT('\n')))
		strSentence = strSentence.Left(strSentence.Length() - 1);
		
	if (strSentence.Length() < 3)
	{
		LibRoadnavDebug0(wxT("SerialIO"), wxT("Length too short after removing CR and LFs"));
		return true;
	}
	
	if (strSentence.GetChar(strSentence.Length() - 3) != wxT('*'))
	{
		LibRoadnavDebug0(wxT("SerialIO"), wxT("Asterisk missing"));
		return true;
	}
	
	if (strSentence.GetChar(0) != wxT('$'))
	{
		LibRoadnavDebug0(wxT("SerialIO"), wxT("Dollar sign missing"));
		return true;
	}
	
	for (iPos = 1; iPos < strSentence.Length() - 3; iPos++)
	{
		uiChecksum ^= strSentence.GetChar(iPos);
		uiChecksum &= 0xff;
	}

	LibRoadnavDebug1(wxT("SerialIO"), wxT("Computed checksum: %02x"), uiChecksum);
	
	if (strSentence.Right(2).Lower() != wxString::Format(wxT("%02x"), uiChecksum))
	{
		LibRoadnavDebug0(wxT("SerialIO"), wxT("Checksum mismatch"));
		return true;
	}

	LibRoadnavDebug0(wxT("SerialIO"), wxT("Checksum is good"));
		
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Tries to get updated data from gps unit.
///
/// \param pGPSEvent		Receives updated gps data.
///
/// \return					Status - either GPSStatusErrorDisableGPS for 
///							serious error, or GPSStatusOK if data was
///							obtained.
///
//////////////////////////////////////////////////////////////////////////////
IGPSInterface::EGPSStatus GPSInterface_Serial::GetData(wxGPSEvent * pGPSEvent)
{
	vector<SSatelliteInfo> vSatelliteInfoTmp;
	wxDateTime tStart = wxDateTime::Now();

	if (!m_bSerialInitialized)
	{
		wxString strSerialInit;
		
		if (InitSerialIO())
		{
			m_strLastError = wxT("Disabling GPS due to error opening serial port: ") + GetLastSerialIOError();
			
			wxThread::Sleep(1000);
			
			ShutdownSerialIO();

			return GPSStatusErrorDisableGPS;
		}
		
		strSerialInit = g_pConfig->Read(wxT("GPSSerialInitString"), wxT(""));
		if (strSerialInit != wxT(""))
			WriteSerial(ProcessEscapeChars(strSerialInit).mb_str(*wxConvCurrent));
		
		m_bSerialInitialized = true;
	}

	while (1)
	{
		if (sizeof(m_szBuffer) - strlen(m_szBuffer) < 192)
		{
			LibRoadnavDebug0(wxT("SerialIO"), wxT("Buffer overrun - dumping oldest 192 bytes"));

			memmove(m_szBuffer, m_szBuffer + 192, strlen(m_szBuffer + 192) + 1);
		}
		
		if (ReadSerial(m_szBuffer + strlen(m_szBuffer), 128) > 0)
		{
			LibRoadnavDebug0(wxT("SerialIO"), wxT("Serial data received"));
			m_tLastSerialActivity = wxDateTime::Now();
		}
		else
		{
			LibRoadnavDebug0(wxT("SerialIO"), wxT("No serial data available - sleeping 0.5 seconds"));
			wxThread::Sleep(500);
		}

		LibRoadnavDebug1(wxT("SerialIO"), wxT("Serial buffer: %hs"), m_szBuffer);
		
		while (1)
		{
			// locate first CR, LF, or NULL
			int iEnd;
			wxString strSentence;
			
			for (iEnd = 0; m_szBuffer[iEnd] && m_szBuffer[iEnd] != '\r' && m_szBuffer[iEnd] != '\n'; iEnd++);

			// stop processing once we run out of CR or LFs
			if (!m_szBuffer[iEnd])
				break;
				
			// isolate sentence
			m_szBuffer[iEnd] = 0;
			strSentence = wxString(m_szBuffer, *wxConvCurrent);

			// remove sentence from m_szBuffer
			memmove(m_szBuffer, m_szBuffer + iEnd + 1, strlen(m_szBuffer + iEnd + 1) + 1);
			
			if (strSentence == wxT("ASTRAL") || strSentence == wxT("EARTHA"))
			{
				char szOut[32];
				
				// Echo DeLorme Tripmate's and Earthmate's hello message
				LibRoadnavDebug1(wxT("SerialIO"), wxT("Echoing hello message: %s"), strSentence.c_str());
				sprintf(szOut, "%s\r\n", strSentence.mb_str(*wxConvCurrent));
				WriteSerial(szOut);
			}
			else if (!VerifyGPSChecksum(strSentence)) // check NMEA validity
			{
				LibRoadnavDebug1(wxT("SerialIO"), wxT("Valid GPS sentence received: %s"), strSentence.c_str());
				m_tLastGPSSentence = wxDateTime::Now();
				
				//////////////////////////////////////////////////////////////////
				// find the start of a position sentence
				//////////////////////////////////////////////////////////////////
				if (strSentence.Left(6) == wxT("$GPGGA"))
				{
					LibRoadnavDebug1(wxT("SerialIO"), wxT("Got $GPGGA sentence: %s"), strSentence.c_str());
						
					//////////////////////////////////////////////////////////
					// actual longitude and latitude
					//////////////////////////////////////////////////////////
					double fLong, fLat;

					//////////////////////////////////////////////////////////
					// true = northern hemisphere
					// false = southern hemisphere
					//////////////////////////////////////////////////////////
					bool bNorth = false;

					//////////////////////////////////////////////////////////
					// true = eastern hemisphere
					// false = western hemisphere
					//////////////////////////////////////////////////////////
					bool bEast = false;

					
					// Fix type
					// 0 = invalid
					// 1 = valid SPS
					// 2 = valid DGPS
					// 3 = valid PPS
					long lFixType;

					// skip over the $GPGGA
					strSentence = strSentence.AfterFirst(wxT(','));
					// skip over the UTC
					strSentence = strSentence.AfterFirst(wxT(','));

					// now we have latitude in ddmm.mmmm format
					strSentence.BeforeFirst(',').ToDouble(&fLat);
					strSentence = strSentence.AfterFirst(wxT(','));

					// North or South
					if (strSentence.Left(1) == wxT("N"))
						bNorth = true;

					strSentence = strSentence.AfterFirst(wxT(','));

					// longitutde
					strSentence.BeforeFirst(',').ToDouble(&fLong);
					strSentence = strSentence.AfterFirst(wxT(','));

					// East or west
					if (strSentence.Left(1) == wxT("E"))
						bEast = true;
					
					strSentence = strSentence.AfterFirst(wxT(','));

					// Position fix
					strSentence.BeforeFirst(',').ToLong(&lFixType);
					strSentence = strSentence.AfterFirst(wxT(','));

					// Number of satellites
					strSentence.BeforeFirst(',').ToLong(&m_nSatellites);
					strSentence = strSentence.AfterFirst(wxT(','));

					// HDOP
					strSentence = strSentence.AfterFirst(wxT(','));

					// Altitude in meters
					//strSentence.BeforeFirst(',').ToDouble(&fAltitude);
					strSentence = strSentence.AfterFirst(wxT(','));

					// process the data

					switch (lFixType)
					{
						case 1:
							m_strFixType = wxT("SPS");
							break;
						case 2:
							m_strFixType = wxT("DGPS");
							break;
						case 3:
							m_strFixType = wxT("PPS");
							break;
						default:
							m_strFixType = wxT("Invalid");
							break;
					}

					if (!bNorth)
						fLat *= -1;
					if (!bEast)
						fLong *= -1;

					// now convert ddmm.mmmm to dd.dddddd
					fLat /= 100;
					fLong /= 100;

					double fFraction;
					int iWhole;

					iWhole = (int) fLong;
					fFraction = fLong - iWhole;
					fLong = iWhole + fFraction / 0.6;

					iWhole = (int) fLat;
					fFraction = fLat - iWhole;
					fLat = iWhole + fFraction / 0.6;
					
					// are they non zero (i.e. good fix)?
					if (fabs(fLong) > 1e-6 && fabs(fLat) > 1e-6)
					{
						m_ptGPS.Set(fLong, fLat);

						LibRoadnavDebug1(wxT("SerialIO"), wxT("Lat and long appear valid - %s"), m_ptGPS.FormatPoint().c_str());

						m_bSeenGPGGA = true;
						m_tLastGPSLock = wxDateTime::Now();
					} // if (fabs(fLong) > 1e-6 && fabs(fLat) > 1e-6)
				} // if (strSentence.Left(6) == wxT("$GPGGA"))


				//////////////////////////////////////////////////////////////////
				// find the start of a course over ground sentence
				//////////////////////////////////////////////////////////////////
				if (strSentence.Left(6) == wxT("$GPRMC"))
				{
					double fLat = 0;
					double fLong = 0;
					bool bNorth = false;
					bool bEast = false;
					
					LibRoadnavDebug1(wxT("SerialIO"), wxT("Got $GPRMC sentence: %s"), strSentence.c_str());

					//////////////////////////////////////////////////////////
					// found what appears to be the right sentence
					//////////////////////////////////////////////////////////

					// skip over the $GPRMC
					strSentence = strSentence.AfterFirst(wxT(','));
					
					// skip over the UTC
					strSentence = strSentence.AfterFirst(wxT(','));
					
					// status
					if (strSentence.Left(1) == wxT("A"))
					{
						if (m_strFixType == wxT("Invalid"))
							m_strFixType = wxT("SPS");
					}
					else if (strSentence.Left(1) == wxT("V"))
					{
						m_strFixType = wxT("Invalid");
					}
					
					strSentence = strSentence.AfterFirst(wxT(','));
					
					// now we have latitude in ddmm.mmmm format
					strSentence.BeforeFirst(',').ToDouble(&fLat);
					strSentence = strSentence.AfterFirst(wxT(','));

					// North or South
					if (strSentence.Left(1) == wxT("N"))
						bNorth = true;

					strSentence = strSentence.AfterFirst(wxT(','));

					// longitutde
					strSentence.BeforeFirst(',').ToDouble(&fLong);
					strSentence = strSentence.AfterFirst(wxT(','));

					// East or west
					if (strSentence.Left(1) == wxT("E"))
						bEast = true;
					
					strSentence = strSentence.AfterFirst(wxT(','));

					if (!bNorth)
						fLat *= -1;
					if (!bEast)
						fLong *= -1;

					// now convert ddmm.mmmm to dd.dddddd
					fLat /= 100;
					fLong /= 100;

					double fFraction;
					int iWhole;

					iWhole = (int) fLong;
					fFraction = fLong - iWhole;
					fLong = iWhole + fFraction / 0.6;

					iWhole = (int) fLat;
					fFraction = fLat - iWhole;
					fLat = iWhole + fFraction / 0.6;
					
					// are they non zero (i.e. good fix)?
					if (fabs(fLong) > 1e-6 && fabs(fLat) > 1e-6)
					{
						m_ptGPS.Set(fLong, fLat);

						LibRoadnavDebug1(wxT("SerialIO"), wxT("Lat and long appear valid - %s"), m_ptGPS.FormatPoint().c_str());

						m_tLastGPSLock = wxDateTime::Now();
					} // if (fabs(fLong) > 1e-6 && fabs(fLat) > 1e-6)

					// speed in knots
					strSentence.BeforeFirst(wxT(',')).ToDouble(&m_fSpeed);
					strSentence = strSentence.AfterFirst(wxT(','));
					// convert to MPH
					m_fSpeed *= 1.151;
					
					// heading
					strSentence.BeforeFirst(wxT(',')).ToDouble(&m_fHeading);
					strSentence = strSentence.AfterFirst(wxT(','));
					
					if ((fabs(m_fSpeed) < 1e-6 && fabs(m_fHeading) < 1e-6) || m_fSpeed < m_fMinimumSpeedForHeading)
					{
						// GPS unit isn't reporting heading, or we're traveling too slow for an accurate heading
						m_fHeading = m_fLastHeading;

						LibRoadnavDebug1(wxT("SerialIO"), wxT("New heading: %.1f (using last heading)"), m_fHeading);
					}
					else
					{
						m_fLastHeading = m_fHeading;

						LibRoadnavDebug1(wxT("SerialIO"), wxT("New heading: %.1f (from sentence)"), m_fHeading);
					}
					
					m_bSeenGPRMC = true;
				} // if (strSentence.Left(6) == wxT("$GPRMC"))
				
				//////////////////////////////////////////////////////////////////
				// find the start of a satellite sentence
				//////////////////////////////////////////////////////////////////
				if (strSentence.Left(6) == wxT("$GPGSV"))
				{
					LibRoadnavDebug1(wxT("SerialIO"), wxT("Got $GPGSV sentence: %s"), strSentence.c_str());

					//////////////////////////////////////////////////////////
					// found what appears to be the right sentence
					//////////////////////////////////////////////////////////
					
					long nMessagesPerSet = 0;
					long lSequenceNumber = 0;
					long lSatellitesVisible = 0;
					
					// skip over the $GPGSV
					strSentence = strSentence.AfterFirst(wxT(','));

					// number of messages in set
					strSentence.BeforeFirst(wxT(',')).ToLong(&nMessagesPerSet);
					strSentence = strSentence.AfterFirst(wxT(','));

					// sequence number
					strSentence.BeforeFirst(wxT(',')).ToLong(&lSequenceNumber);
					strSentence = strSentence.AfterFirst(wxT(','));

					// satellites visible
					strSentence.BeforeFirst(wxT(',')).ToLong(&lSatellitesVisible);
					strSentence = strSentence.AfterFirst(wxT(','));
					
					if (lSequenceNumber == 1)
						vSatelliteInfoTmp.clear();
					
					while (strSentence != wxT(""))
					{
						long lID = -1;
						SSatelliteInfo sInfo;
						wxString strTmp;
						
						sInfo.m_fAzimuth = -1;
						sInfo.m_fElevation = -1;
						sInfo.m_fSNR = 0;

						// ID
						strTmp = strSentence.BeforeFirst(wxT(','));
						if (strTmp != wxT(""))
							strTmp.ToLong(&lID);
						strSentence = strSentence.AfterFirst(wxT(','));
						sInfo.m_iID = lID;
					
						// Elevation
						strTmp = strSentence.BeforeFirst(wxT(','));
						if (strTmp != wxT(""))
							strTmp.ToDouble(&sInfo.m_fElevation);
						strSentence = strSentence.AfterFirst(wxT(','));
					
						// Azimuth
						strTmp = strSentence.BeforeFirst(wxT(','));
						if (strTmp != wxT(""))
							strTmp.ToDouble(&sInfo.m_fAzimuth);
						strSentence = strSentence.AfterFirst(wxT(','));
					
						// SNR
						strTmp = strSentence.BeforeFirst(wxT(','));
						if (strTmp != wxT(""))
							strTmp.ToDouble(&sInfo.m_fSNR);
						strSentence = strSentence.AfterFirst(wxT(','));
					
						// validate info
						if (strSentence != wxT("") &&
							sInfo.m_fAzimuth >= 0 &&
							sInfo.m_fElevation >= 0 &&
							sInfo.m_fSNR >= 0 &&
							sInfo.m_iID > 0)
						{
							vSatelliteInfoTmp.push_back(sInfo);
						}
					}
						
					// check if this is the last message in the set
					if (lSequenceNumber == nMessagesPerSet && nMessagesPerSet > 0)
					{
						unsigned int i;
						
						m_nSatellitesVisible = vSatelliteInfoTmp.size();
						for (i = 0; i < vSatelliteInfoTmp.size(); i++)
							m_arSatellitesVisible[i] = vSatelliteInfoTmp[i];
							
						vSatelliteInfoTmp.clear();
					}
				} // if (strSentence.Left(6) == wxT("$GPGSV"))
			} // if (!VerifyGPSChecksum(strSentence))
		} // while (1)

		LibRoadnavDebug2(wxT("SerialIO"), wxT("bSeenGPGGA = %d, bSeenGPRMC = %d"), m_bSeenGPGGA, m_bSeenGPRMC);
		
		if ((m_bSeenGPGGA || m_bSeenGPRMC) && wxDateTime::Now() - m_tLastGPSLock <= wxTimeSpan(0, 0, 10, 0))
		{
			m_bSeenGPGGA = false;
			m_bSeenGPRMC = false;

			*pGPSEvent = wxGPSEvent(1, 1, 1, m_ptGPS, m_fSpeed, m_fHeading, m_strFixType, m_nSatellites, m_arSatellitesVisible, m_nSatellitesVisible, wxT(""));

			return GPSStatusOK;
		}
		else if (wxDateTime::Now() - m_tLastGPSSentence > wxTimeSpan(0, 0, 10, 0) && 
					wxDateTime::Now() - tStart > wxTimeSpan(0, 0, 5, 0))
		{
			// no GPS unit
			*pGPSEvent = wxGPSEvent(1, 0, 0, Point(0, 0), 0, 0, wxT("None"), 0, NULL, 0, wxT(""));

			return GPSStatusOK;
		}
		else if (wxDateTime::Now() - m_tLastGPSLock > wxTimeSpan(0, 0, 10, 0) &&
					wxDateTime::Now() - tStart > wxTimeSpan(0, 0, 10, 0))
		{
			// no lock
			*pGPSEvent = wxGPSEvent(1, 1, 0, Point(0, 0), 0, 0, wxT("None"), 0, m_arSatellitesVisible, m_nSatellitesVisible, wxT(""));

			return GPSStatusOK;
		}

		// check if thread wants to terminate
		if (m_pThread)
		{
			if (m_pThread->TestDestroy())
			{
				*pGPSEvent = wxGPSEvent(1, 0, 0, Point(0, 0), 0, 0, wxT("None"), 0, NULL, 0, wxT(""));

				return GPSStatusOK;
			}
		}
			
		// check for serial inactivity
		if (wxDateTime::Now() - m_tLastSerialActivity > wxTimeSpan(0, 0, 5, 0))
		{
			wxString strSerialInit;

			LibRoadnavDebug0(wxT("SerialIO"), wxT("GPS timeout. Restarting serial I/O"));

			// no serial activity for 5 seconds .. try reconnecting to serial port
			ShutdownSerialIO();
			InitSerialIO();

			strSerialInit = g_pConfig->Read(wxT("GPSSerialInitString"), wxT(""));
			if (strSerialInit != wxT(""))
				WriteSerial(ProcessEscapeChars(strSerialInit).mb_str(*wxConvCurrent));

			m_tLastSerialActivity = wxDateTime::Now();
		}
	}
	
	wxASSERT(0);

	return GPSStatusOK;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Returns an explanation of the last error.
///
//////////////////////////////////////////////////////////////////////////////
wxString GPSInterface_Serial::GetLastError()
{
	return m_strLastError;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Attempt to autodetect if a GPS unit is attached to a serial port.
///
/// \param pThread			Pointer to wxThread obtain this code is running
///							in.
///
/// \return					Status - either GPSStatusOK if a gps unit
///							was found and a lock was detected, or 
///							GPSStatusAutoDetectionGPSDetectedButNoLock
///							if a gps unit was found, but wasn't locked, or
///							GPSStatusAutoDetectionNoGPSDetected for
///							no gps units.
///
//////////////////////////////////////////////////////////////////////////////
IGPSInterface::EGPSStatus GPSInterface_Serial::AutoDetect(wxThread * pThread)
{
	int iPort;
	int arRates[] = {4800, 9600, 19200, 38400, 57600, 115200, 0};
	int iRate;
	bool bGPSDetected = false;
	
	m_pThread = pThread;

	{	
		// try current settings first
		wxGPSEvent ev;
		if (m_bSerialInitialized)
		{
			ShutdownSerialIO();
			m_bSerialInitialized = false;
		}
		
		if (GetData(&ev) == GPSStatusOK)
		{
			if (ev.m_bLocked)
			{
				m_pThread = NULL;
				return GPSStatusOK;
			}
		}
		
		if (ev.m_bActive)
			bGPSDetected = true;
	}
	
	for (iPort = 0; EnumSerialPort(iPort) != wxT(""); iPort++)
	{
		wxString strPort = EnumSerialPort(iPort);
		
		for (iRate = 0; arRates[iRate]; iRate++)
		{
			wxGPSEvent ev;
			
			g_pConfig->Write(wxT("GPSSerialPort"), strPort);
			g_pConfig->Write(wxT("GPSBaudRate"), arRates[iRate]);
			
			if (m_bSerialInitialized)
			{
				ShutdownSerialIO();
				m_bSerialInitialized = false;
			}
			
			if (GetData(&ev) == GPSStatusOK)
			{
				if (ev.m_bLocked)
				{
					m_pThread = NULL;
					return GPSStatusOK;
				}
			}

			if (ev.m_bActive)
				bGPSDetected = true;
			
			if (pThread && pThread->TestDestroy())
			{
				m_pThread = NULL;
				return GPSStatusAutoDetectionNoGPSDetected;
			}
		}
	}

	m_pThread = NULL;
	
	if (bGPSDetected)
		return GPSStatusAutoDetectionGPSDetectedButNoLock;
		
	return GPSStatusAutoDetectionNoGPSDetected;
}

#endif
