/*
 *  Roadnav
 *  PreferencesPageGeneral.cpp
 *
 *  Copyright (c) 2004 - 2007 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains the general page of the preferences dialog box.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#include <wx/wx.h>

#include <vector>

#include "libroadnav/Map.h"
#include "libroadnav/DownloadFiles.h"
#include "libroadnav/MapSupport.h"

#include "App.h"
#include "PreferencesPageGeneral.h"
#include "SerialIO.h"
#include "wxSkinnableFrame.h"
#include "Support.h"

using namespace std;

enum
{
	idClearCache,
	idClearMaps,
	idEnableExitCommand
};

//////////////////////////////////////////////////////////////////////////////
///
/// PreferencesPageGeneral event table
///
//////////////////////////////////////////////////////////////////////////////
BEGIN_EVENT_TABLE(PreferencesPageGeneral, wxPanel)
	EVT_BUTTON(idClearCache, PreferencesPageGeneral::OnClearCache)
	EVT_BUTTON(idClearMaps, PreferencesPageGeneral::OnClearMaps)
	EVT_CHECKBOX(idEnableExitCommand, PreferencesPageGeneral::OnCheckEnableExitCommand)
END_EVENT_TABLE()

//////////////////////////////////////////////////////////////////////////////
///
/// \brief PreferencesPageGeneral constructor - create controls and initialize them
///
//////////////////////////////////////////////////////////////////////////////
PreferencesPageGeneral::PreferencesPageGeneral(wxWindow *parent)
             : wxPanel(parent, -1, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL)
{
	wxBoxSizer * psizerWnd;
	wxFlexGridSizer * psizerGrid;
	int i;
	
	psizerGrid = new wxFlexGridSizer(2, 5, 10);
	
	//////////////////////////////////////////////////////////////////////////
	// Automatic download maps label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Automatically download maps")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Automatic download maps check box
	//////////////////////////////////////////////////////////////////////////

	m_pctlAutoDownloadMaps = new wxCheckBox(this, -1, wxT(""));

	g_pConfig->Read(wxT("AutoDownloadMaps"), &i, 1);
	m_pctlAutoDownloadMaps->SetValue(i);

	psizerGrid->Add(
		m_pctlAutoDownloadMaps, 
		0,
		wxALIGN_LEFT,
		0
	);

	//////////////////////////////////////////////////////////////////////////
	// Cache downloads label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Cache downloads")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Cache downloads check box
	//////////////////////////////////////////////////////////////////////////

	m_pctlCacheDownloads = new wxCheckBox(this, -1, wxT(""));
	m_pctlCacheDownloads->SetValue(g_pConfig->Read(wxT("CacheDownloads"), (long) 1));

	psizerGrid->Add(
		m_pctlCacheDownloads, 
		0,
		wxALIGN_LEFT,
		0
	);

	{
		//////////////////////////////////////////////////////////////////////////
		// Day Skin Path Label
		//////////////////////////////////////////////////////////////////////////
		psizerGrid->Add(
			new wxStaticText(this, -1, wxT("Daytime Skin")),
			0,
			wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
			0
		);
		
		//////////////////////////////////////////////////////////////////////////
		// Day Skin Path Entry
		//////////////////////////////////////////////////////////////////////////

		vector<wxString> vSkins = EnumerateSkins();

		wxString * strSkins = new wxString[vSkins.size()];
		int nSkins = vSkins.size();
		unsigned int iSkin;

		for (iSkin = 0; iSkin < vSkins.size(); iSkin++)
			strSkins[iSkin] = vSkins[iSkin];

		m_pctlSkinPathDay = new wxComboBox(this, -1, g_pConfig->Read(wxT("SkinPath-Day"), wxT("Default-Day")), wxDefaultPosition, wxSize(200, -1), nSkins, strSkins, wxCB_DROPDOWN | wxCB_SORT);

		delete [] strSkins;

		psizerGrid->Add(
			m_pctlSkinPathDay,
			0,
			wxALIGN_LEFT,
			0
		);
	}
	
	{
		//////////////////////////////////////////////////////////////////////////
		// Night Skin Path Label
		//////////////////////////////////////////////////////////////////////////
		psizerGrid->Add(
			new wxStaticText(this, -1, wxT("Nighttime Skin")),
			0,
			wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
			0
		);
		
		//////////////////////////////////////////////////////////////////////////
		// Night Skin Path Entry
		//////////////////////////////////////////////////////////////////////////

		vector<wxString> vSkins = EnumerateSkins();

		wxString * strSkins = new wxString[vSkins.size()];
		int nSkins = vSkins.size();
		unsigned int iSkin;

		for (iSkin = 0; iSkin < vSkins.size(); iSkin++)
			strSkins[iSkin] = vSkins[iSkin];

		m_pctlSkinPathNight = new wxComboBox(this, -1, g_pConfig->Read(wxT("SkinPath-Night"), wxT("Default-Night")), wxDefaultPosition, wxSize(200, -1), nSkins, strSkins, wxCB_DROPDOWN | wxCB_SORT);

		delete [] strSkins;

		psizerGrid->Add(
			m_pctlSkinPathNight,
			0,
			wxALIGN_LEFT,
			0
		);
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Map Directory Label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Map Directory (blank = default)")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Map Directory Entry
	//////////////////////////////////////////////////////////////////////////
	m_pctlLocalDataDirectory = new wxTextCtrl(this, -1, g_pConfig->Read(wxT("LocalDataDirectory"), wxT("")), wxDefaultPosition, wxSize(250, -1));

	psizerGrid->Add(
		m_pctlLocalDataDirectory,
		0,
		wxALIGN_LEFT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Antialiasing label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Antialiased maps (requires fast CPU)")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Antialiasing check box
	//////////////////////////////////////////////////////////////////////////

	m_pctlAntiAliasedMaps = new wxCheckBox(this, -1, wxT(""));

	g_pConfig->Read(wxT("AntialiasedMaps"), &i, 0);
	m_pctlAntiAliasedMaps->SetValue(i);

	psizerGrid->Add(
		m_pctlAntiAliasedMaps, 
		0,
		wxALIGN_LEFT,
		0
	);

	//////////////////////////////////////////////////////////////////////////
	// Zoom maps Label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Zoom Maps")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Zoom maps Entry
	//////////////////////////////////////////////////////////////////////////

	double fDrawingScaleFactor;

	g_pConfig->Read(wxT("DrawingScaleFactor"), &fDrawingScaleFactor, 1);

	m_pctlZoomMaps = new wxTextCtrl(this, -1, wxString::Format(wxT("%.2f"), fDrawingScaleFactor), wxDefaultPosition, wxSize(250, -1));

	psizerGrid->Add(
		m_pctlZoomMaps,
		0,
		wxALIGN_LEFT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Speed Controlled Zooming Cutoffs Label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Speed Controlled Zooming Cutoffs (mph)")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Speed Controlled Zooming Cutoffs Entry
	//////////////////////////////////////////////////////////////////////////

	wxString strSpeedControlledCutoffs;

	g_pConfig->Read(wxT("SpeedControlledDetailLevelCutoffs"), &strSpeedControlledCutoffs, wxT("25, 50"));

	m_pctlSpeedControlledCutoffs = new wxTextCtrl(this, -1, strSpeedControlledCutoffs, wxDefaultPosition, wxSize(250, -1));

	psizerGrid->Add(
		m_pctlSpeedControlledCutoffs,
		0,
		wxALIGN_LEFT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Home zip code label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Home Zip Code")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Home zip code
	//////////////////////////////////////////////////////////////////////////

   int zipHome;
	g_pConfig->Read(wxT("HomeZipCode"), &zipHome, 0);

	m_pctlHomeZip = new wxTextCtrl(this, -1, wxString::Format(wxT("%05d"), zipHome), wxDefaultPosition, wxSize(250, -1));

	psizerGrid->Add(
		m_pctlHomeZip,
		0,
		wxALIGN_LEFT,
		0
	);

	//////////////////////////////////////////////////////////////////////////
	// Enable exit command label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Enable Exit Command")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Enable exit command checkbox
	//////////////////////////////////////////////////////////////////////////
	m_pctlEnableExitCommand = new wxCheckBox(this, idEnableExitCommand, wxT(""));
	m_pctlEnableExitCommand->SetValue(g_pConfig->Read(wxT("EnableExitCommand"), (long) 0));

	psizerGrid->Add(
		m_pctlEnableExitCommand,
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_LEFT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Exit command label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Exit Command")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Exit command itself
	//////////////////////////////////////////////////////////////////////////
	m_pctlExitCommand = new wxTextCtrl(this, -1, g_pConfig->Read(wxT("ExitCommand"), GetDefaultShutdownCommand()), wxDefaultPosition, wxSize(250, -1));

	psizerGrid->Add(
		m_pctlExitCommand,
		0,
		wxALIGN_LEFT,
		0
	);

	wxCommandEvent ev;
	OnCheckEnableExitCommand(ev);
	
	//////////////////////////////////////////////////////////////////////////
	// Always use advanced waypoint dialog box
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Always Use Advanced Waypoint Dialog Box")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Generate debug log checkbox
	//////////////////////////////////////////////////////////////////////////
	m_pctlAlwaysUseAdvancedWaypointsDialogBox = new wxCheckBox(this, -1, wxT(""));
	m_pctlAlwaysUseAdvancedWaypointsDialogBox->SetValue(g_pConfig->Read(wxT("AlwaysUseAdvancedWaypointsDialog"), (long) 0));

	psizerGrid->Add(
		m_pctlAlwaysUseAdvancedWaypointsDialogBox,
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_LEFT,
		0
	);
	
#ifdef LIBROADNAV_DEBUG
	//////////////////////////////////////////////////////////////////////////
	// Generate debug log label
	//////////////////////////////////////////////////////////////////////////
	psizerGrid->Add(
		new wxStaticText(this, -1, wxT("Generate debug log")),
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT,
		0
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Generate debug log checkbox
	//////////////////////////////////////////////////////////////////////////
	m_pctlGenerateDebugLog = new wxCheckBox(this, -1, wxT(""));
	m_pctlGenerateDebugLog->SetValue(g_pConfig->Read(wxT("GenerateDebugLog"), (long) 0));

	psizerGrid->Add(
		m_pctlGenerateDebugLog,
		0,
		wxALIGN_CENTER_VERTICAL | wxALIGN_LEFT,
		0
	);
	
#endif	

	//////////////////////////////////////////////////////////////////////////
	// Clear cache button
	//////////////////////////////////////////////////////////////////////////
	wxButton * pctlClearCache = new wxButton(this, idClearCache, wxT("Clear Cache"));

	psizerGrid->Add(
		pctlClearCache,
		0,
		wxALIGN_RIGHT,
		0
	);

	//////////////////////////////////////////////////////////////////////////
	// Clear maps button
	//////////////////////////////////////////////////////////////////////////
	wxButton * pctlClearMaps = new wxButton(this, idClearMaps, wxT("Clear Map Data"));

	psizerGrid->Add(
		pctlClearMaps,
		0,
		wxALIGN_LEFT,
		0
	);

	//////////////////////////////////////////////////////////////////////////
	// psizerWnd just adds a border to psizerGrid
	//////////////////////////////////////////////////////////////////////////
	
	psizerWnd = new wxBoxSizer(wxVERTICAL);
	psizerWnd->Add(
		psizerGrid,
		0,
		wxALL | wxGROW,
		10
	);
	
	//////////////////////////////////////////////////////////////////////////
	// Set up the sizer
	//////////////////////////////////////////////////////////////////////////

	m_pctlAutoDownloadMaps->SetFocus();
	psizerWnd->Fit(this);
	SetSizer(psizerWnd);
	Layout();
	psizerWnd->SetSizeHints(this);
	
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Clear cache button pressed - warn the user, then clear the cache.
///
//////////////////////////////////////////////////////////////////////////////
void PreferencesPageGeneral::OnClearCache(wxCommandEvent& event)
{
	if (wxMessageBox(wxT("This will erase all cached downloads. It will not affect the data used to draw the maps. Are you sure you want to continue?"), wxT("Confirm"), wxYES_NO | wxNO_DEFAULT | wxICON_QUESTION, this) == wxYES)
	{
		ClearDownloadCache();
		wxMessageBox(wxT("Cache cleared."), wxT("Information"), wxOK | wxICON_INFORMATION, this);
	}
}

//////////////////////////////////////////////////////////////////////////////
/// 
/// \brief Clear maps button pressed - warn user, then do wipe out compiled maps.
///
//////////////////////////////////////////////////////////////////////////////
void PreferencesPageGeneral::OnClearMaps(wxCommandEvent& event)
{
	if (wxMessageBox(wxT("This will erase all compiled map data. It is recommended you restart Roadnav after performing this operation. Are you sure you want to continue?"), wxT("Confirm"), wxYES_NO | wxNO_DEFAULT | wxICON_QUESTION, this) == wxYES)
	{
		ClearCompiledMapData();
		wxMessageBox(wxT("Map data cleared. Please restart Roadnav."), wxT("Information"), wxOK | wxICON_INFORMATION, this);
	}
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Ok was pressed .. save the settings
///
//////////////////////////////////////////////////////////////////////////////
void PreferencesPageGeneral::OnOk(wxCommandEvent& event)
{
	g_pConfig->Write(wxT("AutoDownloadMaps"), m_pctlAutoDownloadMaps->GetValue());
	g_pConfig->Write(wxT("CacheDownloads"), m_pctlCacheDownloads->GetValue());
	g_pConfig->Write(wxT("AntiAliasedMaps"), m_pctlAntiAliasedMaps->GetValue());
	g_pConfig->Write(wxT("SkinPath-Day"), m_pctlSkinPathDay->GetValue());
	g_pConfig->Write(wxT("SkinPath-Night"), m_pctlSkinPathNight->GetValue());
	g_pConfig->Write(wxT("LocalDataDirectory"), m_pctlLocalDataDirectory->GetValue());
	g_pConfig->Write(wxT("SpeedControlledDetailLevelCutoffs"), m_pctlSpeedControlledCutoffs->GetValue());
	g_pConfig->Write(wxT("EnableExitCommand"), m_pctlEnableExitCommand->GetValue());
	g_pConfig->Write(wxT("ExitCommand"), m_pctlExitCommand->GetValue());
	g_pConfig->Write(wxT("AlwaysUseAdvancedWaypointsDialog"), m_pctlAlwaysUseAdvancedWaypointsDialogBox->GetValue());
#ifdef LIBROADNAV_DEBUG
	long lPreviousDebugSetting = g_pConfig->Read(wxT("GenerateDebugLog"), (long) 0);
	long lNewDebugSetting = m_pctlGenerateDebugLog->GetValue();
	g_pConfig->Write(wxT("GenerateDebugLog"), lNewDebugSetting);
	
	if (lNewDebugSetting && !lPreviousDebugSetting)
	{
		wxMessageBox(wxString::Format(wxT("Debugging is now enabled. Debug messages will be written to %s."), LibRoadnavDebugGetFilename().c_str()),
						wxT("Information"),
						wxOK | wxICON_INFORMATION, 
						this);
	}
	if (!lNewDebugSetting && lPreviousDebugSetting)
	{
		wxMessageBox(wxT("Debugging will be disabled the next time you start Roadnav."),
						wxT("Information"),
						wxOK | wxICON_INFORMATION, 
						this);
	}
#endif

	double fDrawingScaleFactor = 1;

	m_pctlZoomMaps->GetValue().ToDouble(&fDrawingScaleFactor);

	if (fDrawingScaleFactor < 1e-5)
		fDrawingScaleFactor = 1e-5;

	if (fDrawingScaleFactor > 10)
		fDrawingScaleFactor = 10;

	g_pConfig->Write(wxT("DrawingScaleFactor"), fDrawingScaleFactor);

   long lHomeZip;
   m_pctlHomeZip->GetValue().ToLong(&lHomeZip);
   g_pConfig->Write(wxT("HomeZipCode"), lHomeZip);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Enable shutdown command was checked/unchecked
///
//////////////////////////////////////////////////////////////////////////////
void PreferencesPageGeneral::OnCheckEnableExitCommand(wxCommandEvent & event)
{
	m_pctlExitCommand->Enable(m_pctlEnableExitCommand->GetValue());
}

