/*
 *  Roadnav
 *  wxSkinnableFrame.cpp
 *
 *  Copyright (c) 2004 - 2007 Richard L. Lynch <rllynch@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public License
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
//////////////////////////////////////////////////////////////////////////////
/// \file
///
/// Contains the wxSkinnableFrame class and its support code - easily 
/// converts a regular frame into a skinnable one.
///
//////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef _MSC_VER
#pragma warning(disable: 4786)
#pragma warning(disable: 4800)
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <wx/wx.h>
#include <wx/filename.h>
#include <wx/tokenzr.h>

#include "wxSkinnableFrame.h"
#include "libroadnav/MapRepresentations.h"
#include "libroadnav/MapSupport.h"
#include "wxBitmapButtonNoBackground.h"
#include "wxBitmapControl.h"
#include "wxStaticTextNoBackground.h"

#include <vector>

using namespace std;

#define PWS_STYLE(s,a,l)		if (s == wxT(#a)) l |= a;

long ParseWindowStyle(const wxString & str)
{
	long lStyle = 0;
	wxStringTokenizer tkz(str);
	
	while (tkz.HasMoreTokens())
	{
		wxString strTok = tkz.GetNextToken();
		
		PWS_STYLE(strTok, wxSIMPLE_BORDER, lStyle)
		PWS_STYLE(strTok, wxDOUBLE_BORDER, lStyle)
		PWS_STYLE(strTok, wxSUNKEN_BORDER, lStyle)
		PWS_STYLE(strTok, wxRAISED_BORDER, lStyle)
		PWS_STYLE(strTok, wxSTATIC_BORDER, lStyle)
		PWS_STYLE(strTok, wxNO_BORDER, lStyle)
		PWS_STYLE(strTok, wxTRANSPARENT_WINDOW, lStyle)
		PWS_STYLE(strTok, wxTAB_TRAVERSAL, lStyle)
		PWS_STYLE(strTok, wxWANTS_CHARS, lStyle)
		PWS_STYLE(strTok, wxNO_FULL_REPAINT_ON_RESIZE, lStyle)
		PWS_STYLE(strTok, wxVSCROLL, lStyle)
		PWS_STYLE(strTok, wxHSCROLL, lStyle)
		PWS_STYLE(strTok, wxALWAYS_SHOW_SB, lStyle)
		PWS_STYLE(strTok, wxCLIP_CHILDREN, lStyle)
		PWS_STYLE(strTok, wxFULL_REPAINT_ON_RESIZE, lStyle)
		PWS_STYLE(strTok, wxDEFAULT_FRAME_STYLE, lStyle)
		PWS_STYLE(strTok, wxICONIZE, lStyle)
		PWS_STYLE(strTok, wxCAPTION, lStyle)
		PWS_STYLE(strTok, wxMINIMIZE, lStyle)
		PWS_STYLE(strTok, wxMINIMIZE_BOX, lStyle)
		PWS_STYLE(strTok, wxMAXIMIZE, lStyle)
		PWS_STYLE(strTok, wxMAXIMIZE_BOX, lStyle)
		PWS_STYLE(strTok, wxCLOSE_BOX, lStyle)
		PWS_STYLE(strTok, wxSTAY_ON_TOP, lStyle)
		PWS_STYLE(strTok, wxSYSTEM_MENU, lStyle)
		PWS_STYLE(strTok, wxRESIZE_BORDER, lStyle)
		PWS_STYLE(strTok, wxFRAME_TOOL_WINDOW, lStyle)
		PWS_STYLE(strTok, wxFRAME_NO_TASKBAR, lStyle)
		PWS_STYLE(strTok, wxFRAME_FLOAT_ON_PARENT, lStyle)
		PWS_STYLE(strTok, wxFRAME_EX_CONTEXTHELP, lStyle)
		PWS_STYLE(strTok, wxFRAME_SHAPED, lStyle)
		PWS_STYLE(strTok, wxFRAME_EX_METAL, lStyle)
	}
	return lStyle;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Make a vector of the names of the skins in the strDir directory.
///
//////////////////////////////////////////////////////////////////////////////
vector<wxString> EnumerateSkinsOneDir(wxString strDir)
{
	vector<wxString> vRtn;
	wxFileName fnWildcard;
	wxString strFile;

	fnWildcard.AssignDir(strDir);
	fnWildcard.AppendDir(wxT("skins"));

	char szTmp[256];
	
	strcpy(szTmp, fnWildcard.GetFullPath().mb_str(*wxConvCurrent));

	if (!fnWildcard.DirExists())
		return vRtn;

	fnWildcard.SetFullName(wxT("*"));

	strFile = wxFindFirstFile(fnWildcard.GetFullPath(), wxDIR);
	while (!strFile.IsEmpty() )
	{
		wxFileName fnThis = strFile;

		vRtn.push_back(fnThis.GetName());

		strFile = wxFindNextFile();
	}

	return vRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Make a vector of the names of the skins in the common and user
/// directory.
///
//////////////////////////////////////////////////////////////////////////////
vector<wxString> EnumerateSkins()
{
	vector<wxString> vRtn;
	vector<wxString> vTmp;
	vector<wxString>::iterator i;

	vRtn = EnumerateSkinsOneDir(GetCommonDataPath());
	vTmp = EnumerateSkinsOneDir(GetLocalDataPath());

	for (i = vTmp.begin(); i != vTmp.end(); i++)
		vRtn.push_back(*i);

	return vRtn;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Map a skin name to its location.
///
//////////////////////////////////////////////////////////////////////////////
wxString ResolveSkinPath(wxString strName)
{
	wxFileName fn;

	fn = strName;

	if (fn.DirExists())
		return strName;

	fn.AssignDir(GetCommonDataPath());
	fn.AppendDir(wxT("skins"));
	fn.AppendDir(strName);

	if (fn.DirExists())
		return fn.GetFullPath();

	fn.AssignDir(GetLocalDataPath());
	fn.AppendDir(wxT("skins"));
	fn.AppendDir(strName);

	if (fn.DirExists())
		return fn.GetFullPath();

	return wxT("");
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Determine the amount of a space a multiline string would take up
/// on a certain device context.
///
/// strText is the text whose size is being measured
///
/// dc is the device context
///
/// w and h return the width and height of strText on the dc device context
///
//////////////////////////////////////////////////////////////////////////////
void GetTextExtentMultiline(wxDC * dc, wxString strText, wxCoord * w, wxCoord * h)
{
	wxString strLine;
	vector<wxString> vLines;
	int x, y;
	int iLineHeight;
	unsigned int iLine;	
	
	*w = 0;
	*h = 0;

	if (strText == wxT(""))
		return;
			
	strLine = strText.BeforeFirst('\n');
	strText = strText.AfterFirst('\n');
	
	while (strLine != wxT(""))
	{
		vLines.push_back(strLine);

		strLine = strText.BeforeFirst('\n');
		strText = strText.AfterFirst('\n');
	}

	iLineHeight = 0;
	
	for (iLine = 0; iLine < vLines.size(); iLine++)
	{
		dc->GetTextExtent(vLines[iLine], &x, &y);
		
		if (x > *w)
			*w = x;
		if (y > iLineHeight)
			iLineHeight = y;
	}
	
	*h = iLineHeight * vLines.size();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Loads a font from an XML element
///
//////////////////////////////////////////////////////////////////////////////
wxFont LoadFont(TiXmlElement * root)
{
	int iDefaultFontSize = -1;
	wxString strDefaultFontName;
	wxFont fntRtn;
	TiXmlElement * element;

	element = root->FirstChildElement();

	while (element)
	{
		const char * pszName = element->Value();
		const char * pszContents = NULL;
		TiXmlNode * node;

		for (node = element->FirstChild(); node && node->Type() != TiXmlNode::TEXT; node = node->NextSibling());
		if (node)
			pszContents = node->Value();

		if (!strcmp(pszName, "size") && pszContents)
			iDefaultFontSize = atoi(pszContents);
			
		if (!strcmp(pszName, "name") && pszContents)
			strDefaultFontName = wxString(pszContents, *wxConvCurrent);
				
		element = element->NextSiblingElement();
	}
	
	fntRtn = *wxNORMAL_FONT;
	
	if (strDefaultFontName != wxT(""))
		fntRtn.SetFaceName(strDefaultFontName);
		
	if (iDefaultFontSize > 0)
		fntRtn.SetPointSize(iDefaultFontSize);
		
	return fntRtn;
}


//////////////////////////////////////////////////////////////////////////////
///
/// \brief wxSkinnableFrame_ProportionalCoordinate constructor - initialize everything to zero
///
//////////////////////////////////////////////////////////////////////////////
wxSkinnableFrame_ProportionalCoordinate::wxSkinnableFrame_ProportionalCoordinate()
{
	m_fMultiplier = 0;
	m_iOffset1 = 0;
	m_iOffset2 = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Loads a proportional position from an XML element
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame_ProportionalCoordinate::Load(TiXmlElement * root)
{
	root->QueryIntAttribute("offset1", &m_iOffset1);
	root->QueryIntAttribute("offset2", &m_iOffset2);
	root->QueryDoubleAttribute("multiplier", &m_fMultiplier);

}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Loads a pair of coordinates from an XML element
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame_ProportionalPoint::Load(TiXmlElement * root)
{
	TiXmlElement * element;

	element = root->FirstChildElement("x");

	if (!element)
		return;

	m_cX.Load(element);


	element = root->FirstChildElement("y");

	if (!element)
		return;
	m_cY.Load(element);


	return;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Resolve this coordinate given the size of the control
///
//////////////////////////////////////////////////////////////////////////////
int wxSkinnableFrame_ProportionalCoordinate::Resolve(int iMax)
{
	return (int) (m_iOffset1 + (m_iOffset2 + iMax) * m_fMultiplier);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Resolve this point given the size of the control
///
//////////////////////////////////////////////////////////////////////////////
wxPoint wxSkinnableFrame_ProportionalPoint::Resolve(wxPoint cMax)
{
	return wxPoint(m_cX.Resolve(cMax.x), m_cY.Resolve(cMax.y));
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief wxSkinnableFrame_LayoutControl constructor
///
//////////////////////////////////////////////////////////////////////////////
wxSkinnableFrame_LayoutControl::wxSkinnableFrame_LayoutControl()
{
	m_cCenterWidth = -1;

	m_iShowMinHeight = 0;
	m_iShowMinWidth = 0;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Position this control according to m_coTopLeft and m_coBottomRight
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame_LayoutControl::Position(wxSize szParent)
{
	wxPoint ptTopLeft;
	wxPoint ptBottomRight;
	bool bWantVisible;

	ptTopLeft = m_coTopLeft.Resolve(wxPoint(szParent.x, szParent.y));
	ptBottomRight = m_coBottomRight.Resolve(wxPoint(szParent.x, szParent.y));

	if (ptTopLeft.x >= szParent.x)
		ptTopLeft.x = szParent.x - 1;
	if (ptTopLeft.y >= szParent.y)
		ptTopLeft.y = szParent.y - 1;

	if (ptBottomRight.x >= szParent.x)
		ptBottomRight.x = szParent.x - 1;
	if (ptBottomRight.y >= szParent.y)
		ptBottomRight.y = szParent.y - 1;
		
	if (m_cCenterWidth >= 0)
	{
		ptTopLeft.x = (ptBottomRight.x + ptTopLeft.x) / 2 - m_cCenterWidth / 2;
		ptBottomRight.x = (ptBottomRight.x + ptTopLeft.x) / 2 + m_cCenterWidth / 2 + 5;
	}

	m_pControl->SetSize(ptTopLeft.x, ptTopLeft.y, ptBottomRight.x - ptTopLeft.x, ptBottomRight.y - ptTopLeft.y);

	bWantVisible = true;
	
	if (szParent.y < m_iShowMinHeight || szParent.x < m_iShowMinWidth)
		bWantVisible = false;

	if (m_pControl->IsShown() != bWantVisible)
		m_pControl->Show(bWantVisible);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Event table for wxSkinnableTable
///
//////////////////////////////////////////////////////////////////////////////
BEGIN_EVENT_TABLE(wxSkinnableFrame, wxFrame)
	EVT_SIZE(wxSkinnableFrame::OnSize)
	EVT_TIMER(TIMER_REFRESH, wxSkinnableFrame::OnTimerRefresh)
END_EVENT_TABLE()

using namespace std;

//////////////////////////////////////////////////////////////////////////////
///
/// \brief wxSkinnableFrame constructor - just calls base constructor
///
//////////////////////////////////////////////////////////////////////////////
wxSkinnableFrame::wxSkinnableFrame(	wxWindow* parent, 
									wxWindowID id, 
									const wxString& title, 
									const wxPoint& pos, 
									const wxSize& size, 
									long style, 
									const wxString& name)
									: wxFrame(parent, id, title, pos, size, style, name)
{
	m_pTimerRefresh = new wxTimer(this, TIMER_REFRESH);
}

wxSkinnableFrame::~wxSkinnableFrame()
{
	delete m_pTimerRefresh;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Given a filename to a skin data file, generate a complete pathname
///
//////////////////////////////////////////////////////////////////////////////
wxString wxSkinnableFrame::GetDataFilename(wxString strName)
{
	wxFileName fn;

	fn.AssignDir(m_strSkinPath);
	fn.SetFullName(strName);

	return fn.GetFullPath();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Apply generic window attributes
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::ApplyWindowCommonAttributes(TiXmlElement * root, wxWindow * pWindow)
{
	const char * pszTmp;
	int iX = -1;
	int iY = -1;
	int iW = -1;
	int iH = -1;
	
	pszTmp = root->Attribute("style");
	if (pszTmp)
		pWindow->SetWindowStyle(ParseWindowStyle(wxString(pszTmp, *wxConvCurrent)));
	
	pszTmp = root->Attribute("x");
	if (pszTmp)
		iX = atoi(pszTmp);
	
	pszTmp = root->Attribute("y");
	if (pszTmp)
		iY = atoi(pszTmp);
	
	pszTmp = root->Attribute("w");
	if (pszTmp)
		iW = atoi(pszTmp);
	
	pszTmp = root->Attribute("h");
	if (pszTmp)
		iH = atoi(pszTmp);
	
	if (iX >= 0 && iY >= 0 && iW >= -1 && iH >= -1)
	{
		SetSize(iX, iY, iW, iH);
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load attributes from an XML element that are common to all of the
/// controls like its region, foreground/background color, etc.
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::ApplyControlCommonAttributes(TiXmlElement * root, wxSkinnableFrame_LayoutControl * psLayoutControl, wxControl * pControl)
{
	TiXmlElement * element;
	TiXmlElement * child;
	const char * pszTmp;

	element = root->FirstChildElement("region");
	if (!element)
		return true;

	// top left
	child = element->FirstChildElement("topleft");

	if (!child)
		return true;

	psLayoutControl->m_coTopLeft.Load(child);


	// bottom right

	child = element->FirstChildElement("bottomright");

	if (!child)
		return true;

	psLayoutControl->m_coBottomRight.Load(child);

	// font
	child = root->FirstChildElement("font");
	if (child)
	{
		wxFont fntThis = LoadFont(child);
		pControl->SetFont(fntThis);
	}

	pszTmp = root->Attribute("bgcolor");
	if (pszTmp)
		pControl->SetBackgroundColour(ParseColor(wxString(pszTmp, *wxConvCurrent)));

	pszTmp = root->Attribute("fgcolor");
	if (pszTmp)
		pControl->SetForegroundColour(ParseColor(wxString(pszTmp, *wxConvCurrent)));

	pszTmp = root->Attribute("showminheight");
	if (pszTmp)
		psLayoutControl->m_iShowMinHeight = atoi(pszTmp);

	pszTmp = root->Attribute("showminwidth");
	if (pszTmp)
		psLayoutControl->m_iShowMinWidth = atoi(pszTmp);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the control element in layout
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinLayoutControl(TiXmlElement * root)
{
	const char * pszAttributeName;
	wxSkinnableFrame_LayoutControl sLayoutControl;

	// name
	pszAttributeName = root->Attribute("name");

	if (!pszAttributeName)
		return true;

	// request the derived class create the control now
	// so it appears in the proper place in the z-order
	CreateNamedControl(wxString(pszAttributeName, *wxConvCurrent));

	// assign the control itself to sLayoutControl
	sLayoutControl.m_pControl = m_mapUnassignedControls[pszAttributeName];

	sLayoutControl.m_eType = wxSkinnableFrame_LayoutControl::ctControl;
	sLayoutControl.m_strName = wxString(pszAttributeName, *wxConvCurrent);

	if (!sLayoutControl.m_pControl)
		return true;

	if (ApplyControlCommonAttributes(root, &sLayoutControl, sLayoutControl.m_pControl))
		return true;

	m_mapUnassignedControls[pszAttributeName] = NULL;

	// store in vector
	m_vPositionedControls.push_back(sLayoutControl);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the button element in layout
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinLayoutButton(TiXmlElement * root)
{
	wxBitmap bmpDefault;
	wxBitmap bmpDepressed;
	wxBitmap bmpActive;
	const char * pszDefaultFilename;
	const char * pszDepressedFilename;
	const char * pszActiveFilename;
	const char * pszAction;
	const char * pszActiveKey;
	wxBitmapButtonNoBackground * pCtl;
	wxSkinnableFrame_LayoutControl sLayoutControl;
	int iID;

	pszDefaultFilename = root->Attribute("defaultimage");
	pszDepressedFilename = root->Attribute("depressedimage");
	pszActiveFilename = root->Attribute("activeimage");
	pszActiveKey = root->Attribute("activekey");
	pszAction = root->Attribute("action");

	if (!pszDefaultFilename || !pszDepressedFilename || !pszAction)
		return true;

	if (!bmpDefault.LoadFile(GetDataFilename(wxString(pszDefaultFilename, *wxConvCurrent)), wxBITMAP_TYPE_ANY))
		return true;

	if (!bmpDepressed.LoadFile(GetDataFilename(wxString(pszDepressedFilename, *wxConvCurrent)), wxBITMAP_TYPE_ANY))
		return true;

	if (pszActiveFilename && pszActiveKey)
	{
		if (!bmpActive.LoadFile(GetDataFilename(wxString(pszActiveFilename, *wxConvCurrent)), wxBITMAP_TYPE_ANY))
			return true;
	}

	iID = MapNameToID(wxString(pszAction, *wxConvCurrent));

	if (iID < 0)
		return true;
	
	// control itself
	pCtl = new wxBitmapButtonNoBackground(this, iID, bmpDefault, wxPoint(0, 0), wxSize(1, 1), 0);
	pCtl->SetBitmapSelected(bmpDepressed);
	
	if (pszActiveFilename && pszActiveKey)
		pCtl->SetBitmapActive(bmpActive);

	sLayoutControl.m_pControl = pCtl;
	sLayoutControl.m_eType = wxSkinnableFrame_LayoutControl::ctButton;
	sLayoutControl.m_strName = wxString(pszAction, *wxConvCurrent);

	if (pszActiveFilename && pszActiveKey)
		sLayoutControl.m_strActiveKey = wxString(pszActiveKey, *wxConvCurrent);

	if (ApplyControlCommonAttributes(root, &sLayoutControl, sLayoutControl.m_pControl))
	{
		delete pCtl;
		return true;
	}

	// store in vector
	m_vPositionedControls.push_back(sLayoutControl);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the button element in layout
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinLayoutImage(TiXmlElement * root)
{
	wxBitmap bmpDefault;
	wxBitmap bmpDepressed;
	const char * pszDefaultFilename;
#if defined(__WINDOWS__)
	wxBitmapControl * pCtl;
#else
	wxStaticBitmap * pCtl;
#endif
	wxSkinnableFrame_LayoutControl sLayoutControl;

	pszDefaultFilename = root->Attribute("filename");

	if (!pszDefaultFilename)
		return true;

	if (!bmpDefault.LoadFile(GetDataFilename(wxString(pszDefaultFilename, *wxConvCurrent)), wxBITMAP_TYPE_ANY))
		return true;

	// control itself
#if defined(__WINDOWS__)
	pCtl = new wxBitmapControl(this, -1, bmpDefault, wxPoint(0, 0), wxSize(1, 1), 0);
	pCtl->Enable(false);
#else
	pCtl = new wxStaticBitmap(this, -1, bmpDefault, wxPoint(0, 0), wxSize(1, 1), 0);
#endif

	sLayoutControl.m_pControl = pCtl;
	sLayoutControl.m_eType = wxSkinnableFrame_LayoutControl::ctImage;
	sLayoutControl.m_strName = wxString(pszDefaultFilename, *wxConvCurrent);

	if (ApplyControlCommonAttributes(root, &sLayoutControl, sLayoutControl.m_pControl))
	{
		delete pCtl;
		return true;
	}

	// store in vector
	m_vPositionedControls.push_back(sLayoutControl);

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the text element in layout
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinLayoutText(TiXmlElement * root)
{
	wxBitmap bmpDefault;
	wxBitmap bmpDepressed;
	const char * pszName;
	const char * pszLabel;
	const char * pszPrepend;
	const char * pszAppend;
	wxStaticTextNoBackground * pCtl;
	wxSkinnableFrame_LayoutControl sLayoutControl;

	pszName = root->Attribute("name");
	pszLabel = root->Attribute("label");
	pszPrepend = root->Attribute("prepend");
	pszAppend = root->Attribute("append");

	if (!pszName)
		pszName = "";

	if (!pszLabel)
		pszLabel = "";
	
	if (!pszPrepend)
		pszPrepend = "";

	if (!pszAppend)
		pszAppend = "";

	// control itself
	pCtl = new wxStaticTextNoBackground(this, -1, wxT("xxx"), wxPoint(0, 0), wxSize(1, 1), wxALIGN_LEFT | wxST_NO_AUTORESIZE);
	sLayoutControl.m_pControl = pCtl;
	sLayoutControl.m_eType = wxSkinnableFrame_LayoutControl::ctText;
	sLayoutControl.m_strName = wxString(pszName, *wxConvCurrent);
	sLayoutControl.m_strPrepend = wxString(pszPrepend, *wxConvCurrent);
	sLayoutControl.m_strAppend = wxString(pszAppend, *wxConvCurrent);

	if (ApplyControlCommonAttributes(root, &sLayoutControl, sLayoutControl.m_pControl))
	{
		delete pCtl;
		return true;
	}

	// store in vector
	m_vPositionedControls.push_back(sLayoutControl);

	SetNamedText(wxString(pszName, *wxConvCurrent), wxString(pszLabel, *wxConvCurrent), false);
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the mainwindow element
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinMainWindow(TiXmlElement * root)
{
	ApplyWindowCommonAttributes(root, this);
	return false;	
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the layout element
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinLayout(TiXmlElement * root)
{
	TiXmlElement * element;

	element = root->FirstChildElement();

	while (element)
	{
		const char * pszName = element->Value();

		if (!strcmp(pszName, "control"))
			LoadSkinLayoutControl(element);

		if (!strcmp(pszName, "button"))
			LoadSkinLayoutButton(element);
		
		if (!strcmp(pszName, "image"))
			LoadSkinLayoutImage(element);

		if (!strcmp(pszName, "text"))
			LoadSkinLayoutText(element);
		
		element = element->NextSiblingElement();
	}
	
	return true;	
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Process the global element. Contains various application wide
/// settings like min window size, default font size, etc.
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkinGlobal(TiXmlElement * root)
{
	TiXmlElement * element;
	
	element = root->FirstChildElement();

	while (element)
	{
		const char * pszName = element->Value();

		if (!strcmp(pszName, "font"))
		{
			m_fntDefault = LoadFont(element);
			SetFont(m_fntDefault);
		}
		
		if (!strcmp(pszName, "minsize"))
		{
			int iWidth = 0;
			int iHeight = 0;
			const char * pszTmp;
			
			pszTmp = element->Attribute("width");
			if (pszTmp)
				iWidth = atoi(pszTmp);

			pszTmp = element->Attribute("height");
			if (pszTmp)
				iHeight = atoi(pszTmp);
				
			if (iWidth && iHeight)
			{
				wxSize szThis;
				wxSize szMargin;
				bool bSizeChanged = false;
				
				// figure out how big the status bar, menu bar, etc. are
				szMargin = GetSize() - GetClientSize();
				
				// convert minsize from client size to window size
				iWidth += szMargin.x;
				iHeight += szMargin.y;

				// set min size
				SetSizeHints(iWidth, iHeight);

				// grow window if necessary
				szThis = GetSize();

				if (szThis.x < iWidth)
				{
					szThis.x = iWidth;
					bSizeChanged = true;
				}

				if (szThis.y < iHeight)
				{
					szThis.y = iHeight;
					bSizeChanged = true;
				}

				if (bSizeChanged)
					SetSize(szThis);

			}
		}
				
		element = element->NextSiblingElement();
	}
	
	return true;	
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Load the skin in the specified directory
///
//////////////////////////////////////////////////////////////////////////////
bool wxSkinnableFrame::LoadSkin(wxString strPath, bool bIsFilename)
{
	wxMutexLocker lock(m_mxSkin);
	TiXmlDocument doc;
	
	if (m_strCurrentSkin == strPath)
		return false;
	
	UnloadSkin();

	if (bIsFilename)
	{
		m_strSkinPath = ResolveSkinPath(strPath);

		doc.SetValue(GetDataFilename(wxT("skin.xml")).mb_str(*wxConvCurrent));
		doc.LoadFile();
	}
	else
	{
		m_strSkinPath = wxT("");

		doc.Parse(strPath.mb_str(*wxConvCurrent));
	}

	TiXmlElement * element;


	element = doc.RootElement();

	if (!element)
		return true;

	element = element->FirstChildElement();

	while (element)
	{
		const char * pszName = element->Value();

		if (!strcmp(pszName, "layout"))
			LoadSkinLayout(element);

		if (!strcmp(pszName, "global"))
			LoadSkinGlobal(element);

		if (!strcmp(pszName, "mainwindow"))
			LoadSkinMainWindow(element);

		element = element->NextSiblingElement();
	}
	
	map<std::string,wxControl *>::iterator i;
	for (i = m_mapUnassignedControls.begin(); i != m_mapUnassignedControls.end(); i++)
	{
		if ((*i).second)
			(*i).second->Show(false);
	}
	
	wxTimerEvent event;
	OnTimerRefresh(event);
	
	m_strCurrentSkin = strPath;

	return false;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Handle the window being resized
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::OnSize(wxSizeEvent & evEvent)
{
	if (m_vPositionedControls.size() == 0)
	{
		wxFrame::OnSize(evEvent);
		return;
	}

	m_pTimerRefresh->Start(100, TRUE);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Add a named control - one that can be positioned by the skin
/// using the control element.
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::AddNamedControl(wxControl * pCtl, wxString strName)
{
	m_mapUnassignedControls[static_cast<const char *>(strName.mb_str(*wxConvCurrent))] = pCtl;
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief This function should be overriden by derived classes to produce
/// the class called strName, and then call AddNamedControl.
///
/// Such controls cannot be created before LoadSkin, since they'd appear
/// in the wrong z-order (beneath all of the skin controls).
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::CreateNamedControl(wxString strName)
{
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief This function should be overriden by derived classes to
/// delete the object called strName that was created in CreateNamedControl
///
/// If the control was created on the stack and added during the frame
/// constructor, AddNamedControl should be called again.
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::DeleteNamedControl(wxString strName, wxControl * pControl)
{
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Update the text in the text field called strName with text 
/// strNewLabel
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::SetNamedText(wxString strName, wxString strNewLabel, bool bNeedLock)
{
	if (bNeedLock)
		m_mxSkin.Lock();

	std::vector<wxSkinnableFrame_LayoutControl>::iterator i;

	for (i = m_vPositionedControls.begin(); i != m_vPositionedControls.end(); i++)
	{
		if ((*i).m_strName == strName && (*i).m_eType == wxSkinnableFrame_LayoutControl::ctText)
		{
			wxString strCompleteLabel = (*i).m_strPrepend + strNewLabel + (*i).m_strAppend;

			if ((*i).m_pControl->GetLabel() != strCompleteLabel)
			{
				// restore the static text to its original size
				wxSize szCli = GetClientSize();
				wxScreenDC dcScr;
				wxCoord w, h;
				
				dcScr.SetFont((*i).m_pControl->GetFont());
				GetTextExtentMultiline(&dcScr, strCompleteLabel, &w, &h);

				(*i).m_cCenterWidth = w;

				(*i).Position(szCli);
						
				(*i).m_pControl->SetLabel(strCompleteLabel);
			}
		}
	}

	if (bNeedLock)
		m_mxSkin.Unlock();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Update the active state on controls with activation key
/// strActiveKey
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::SetActive(wxString strActiveKey, bool bActive, bool bNeedLock)
{
	if (bNeedLock)
		m_mxSkin.Lock();

	std::vector<wxSkinnableFrame_LayoutControl>::iterator i;

	for (i = m_vPositionedControls.begin(); i != m_vPositionedControls.end(); i++)
	{
		if ((*i).m_strActiveKey == strActiveKey && (*i).m_eType == wxSkinnableFrame_LayoutControl::ctButton)
		{
			wxBitmapButtonNoBackground * pCtl = wxDynamicCast((*i).m_pControl, wxBitmapButtonNoBackground);
			
			if (pCtl)
				pCtl->SetActive(bActive);
		}
	}

	if (bNeedLock)
		m_mxSkin.Unlock();
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Unload the skin
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::UnloadSkin()
{
	vector<wxSkinnableFrame_LayoutControl>::iterator i;

	for (i = m_vPositionedControls.begin(); i != m_vPositionedControls.end(); i++)
	{
		if ((*i).m_eType == wxSkinnableFrame_LayoutControl::ctControl)
			DeleteNamedControl((*i).m_strName, (*i).m_pControl);
		else
			delete (*i).m_pControl;
	}

	m_vPositionedControls.clear();
	m_strSkinPath = wxT("");
	m_fntDefault = *wxNORMAL_FONT;

	SetFont(m_fntDefault);
}

//////////////////////////////////////////////////////////////////////////////
///
/// \brief Timer event triggered 100ms after the user stops resizing the
/// window. May also schedule itself to be triggered later if it can't get
/// a lock on m_mxSkin.
///
//////////////////////////////////////////////////////////////////////////////
void wxSkinnableFrame::OnTimerRefresh(wxTimerEvent& event)
{
	// somehow, there's potential for a deadlock here if we just use a regular
	// mutex locker. Instead, try to get the lock, and if it's not available,
	// schedule a timer to try again later.
	if (m_mxSkin.TryLock() != wxMUTEX_NO_ERROR)
	{
		m_pTimerRefresh->Start(1, TRUE);
		return;
	}

	vector<wxSkinnableFrame_LayoutControl>::iterator i;
	wxSize szCli = GetClientSize();

	for (i = m_vPositionedControls.begin(); i != m_vPositionedControls.end(); i++)
		(*i).Position(szCli);
		
	Refresh();
	
	m_mxSkin.Unlock();
}
